#include <qobject.h>
#include <qstring.h>
#include <qtextedit.h>
#include <qfile.h>
#include <qtextstream.h>
#include <stdio.h>
#include "table.h"
#include "custom.h"
#include "file_vf.h"
#include "math_p4.h"
#include "p4application.h"

/*
    This file contains the code to read the information from reduce/maple.
    The relevant structures for saddles etc are set up here.
*/

extern void delete_term1( P4POLYNOM1 );
extern void delete_term2( P4POLYNOM2 );
extern void delete_term3( P4POLYNOM3 );

extern void SetP4WindowTitle( QWidget *, QString );

extern double eval_term2( P4POLYNOM2, double * );

extern const char * DumpPoly1( P4POLYNOM1 f, const char * x );
extern const char * DumpPoly2( P4POLYNOM2 f, const char * x, const char * y );
extern const char * DumpPoly3( P4POLYNOM3 f, const char * x, const char * y, const char * z );


// -----------------------------------------------------------------------
//                              QVFStudy CONSTRUCTOR
// -----------------------------------------------------------------------

QVFStudy::QVFStudy()
{
    // initialize vector field structures:

    f_vec_field[0] = NULL;      f_vec_field[1] = NULL;
    vec_field_U1[0] = NULL;     vec_field_U1[1] = NULL;
    vec_field_U2[0] = NULL;     vec_field_U2[1] = NULL;
    vec_field_V1[0] = NULL;     vec_field_V1[1] = NULL;
    vec_field_V2[0] = NULL;     vec_field_V2[1] = NULL;
    vec_field_C[0] = NULL;      vec_field_C[1] = NULL;

    // initialize singular points structures:

    first_saddle_point = NULL;
    first_se_point = NULL;
    first_node_point = NULL;
    first_sf_point = NULL;
    first_wf_point = NULL;
    first_de_point = NULL;

    // initialize GCF:

    gcf = NULL;
    gcf_U1 = NULL;
    gcf_U2 = NULL;
    gcf_V1 = NULL;
    gcf_V2 = NULL;
    gcf_C = NULL;
    gcf_points = NULL;

    // initialize limit cycles & orbits

    first_lim_cycle = NULL;
    first_orbit = NULL;
    current_orbit = NULL;

    // initialize others

    xmin = -1.0;
    xmax = 1.0;
    ymin = -1.0;
    ymax = 1.0;
    p = 1;
    q = 1;
    typeofstudy = TYPEOFSTUDY_ALL;
    singinf = false;
    VFResults.dir_vec_field = 1;

    // initialize parameters

    config_lc_value = DEFAULT_LCORBITS;         // number of orbits in the limit cycle window
    config_lc_numpoints = DEFAULT_LCPOINTS;     // number of points in the limit cycle window
    config_hma = DEFAULT_HMA;                   // maximum step size
    config_hmi = DEFAULT_HMI;                   // minimum step size
    config_step = DEFAULT_STEPSIZE;             // step size
    config_currentstep = DEFAULT_STEPSIZE;      // current step size (during integration)
    config_tolerance = DEFAULT_TOLERANCE;       // tolerance
    config_projection = DEFAULT_PROJECTION;     // projection in the case of Poincare sphere
    config_intpoints = DEFAULT_INTPOINTS;       // number of points to integrate
    config_dashes = DEFAULT_LINESTYLE;          // line style (dashes or points)
    config_kindvf = DEFAULT_INTCONFIG;
}


// -----------------------------------------------------------------------
//                              QVFStudy DESTRUCTOR
// -----------------------------------------------------------------------

QVFStudy::~QVFStudy()
{
    DeleteVF();
}

// -----------------------------------------------------------------------
//                          QVFStudy::DeleteVF
// -----------------------------------------------------------------------

void QVFStudy::DeleteVF()
{
    // Delete Vector Field

    delete_term2( f_vec_field[0] );     delete_term2( f_vec_field[1] );
    delete_term2( vec_field_U1[0] );    delete_term2( vec_field_U1[1] );
    delete_term2( vec_field_U2[0] );    delete_term2( vec_field_U2[1] );
    delete_term2( vec_field_V1[0] );    delete_term2( vec_field_V1[1] );
    delete_term2( vec_field_V2[0] );    delete_term2( vec_field_V2[1] );
    delete_term3( vec_field_C[0] );     delete_term3( vec_field_C[1] );

    f_vec_field[0] = NULL;      f_vec_field[1] = NULL;
    vec_field_U1[0] = NULL;     vec_field_U1[1] = NULL;
    vec_field_U2[0] = NULL;     vec_field_U2[1] = NULL;
    vec_field_V1[0] = NULL;     vec_field_V1[1] = NULL;
    vec_field_V2[0] = NULL;     vec_field_V2[1] = NULL;
    vec_field_C[0] = NULL;      vec_field_C[1] = NULL;

    // Delete singular points

    DeleteSaddle( first_saddle_point );
    DeleteSemiElementary( first_se_point );
    DeleteNode( first_node_point );
    DeleteDegenerate( first_de_point );
    DeleteStrongFocus( first_sf_point );
    DeleteWeakFocus( first_wf_point );

    first_saddle_point = NULL;
    first_se_point = NULL;
    first_node_point = NULL;
    first_sf_point = NULL;
    first_wf_point = NULL;
    first_de_point = NULL;

    // Delete GCF:

    delete_term2( gcf );
    delete_term2( gcf_U1 );
    delete_term2( gcf_U2 );
    delete_term2( gcf_V1 );
    delete_term2( gcf_V2 );
    delete_term3( gcf_C );
    DeleteOrbitPoint( gcf_points );

    gcf = NULL;
    gcf_U1 = NULL;
    gcf_U2 = NULL;
    gcf_V1 = NULL;
    gcf_V2 = NULL;
    gcf_C = NULL;
    gcf_points = NULL;

    // Delete all orbits

    DeleteOrbit( first_orbit );
    first_orbit = NULL;
    current_orbit = NULL;

    // Delete limit cycles

    DeleteLimitCycle( first_lim_cycle );
    first_lim_cycle = NULL;

    // reset others

    xmin = -1.0;
    xmax = 1.0;
    ymin = -1.0;
    ymax = 1.0;
    p = 1;
    q = 1;
    typeofstudy = TYPEOFSTUDY_ALL;
    singinf = false;
    VFResults.dir_vec_field = 1;
}

// -----------------------------------------------------------------------
//                          QVFStudy::DeleteSaddle
// -----------------------------------------------------------------------

void QVFStudy::DeleteSaddle( struct saddle * p )
{
    struct saddle * q;

    while( p != NULL )
    {
        q = p;
        p = p->next_saddle;
    
        delete_term2( q->vector_field[0] );
        delete_term2( q->vector_field[1] );
        if( q->dummy )
            DeleteSeparatrices( q->separatrices );  // separatrices may be NULL
        free(q);
    }
}

// -----------------------------------------------------------------------
//                      QVFStudy::DeleteSemiElementary
// -----------------------------------------------------------------------

void QVFStudy::DeleteSemiElementary( struct semi_elementary * p )
{
    struct semi_elementary * q;

    while( p != NULL )
    {
        q = p;
        p = p->next_se;
    
        delete_term2( q->vector_field[0] );
        delete_term2( q->vector_field[1] );
        if( q->dummy )
            DeleteSeparatrices( q->separatrices );  // separatrices may be NULL
        free(q);
    }
}

// -----------------------------------------------------------------------
//                          QVFStudy::DeleteNode
// -----------------------------------------------------------------------

void QVFStudy::DeleteNode( struct node * p )
{
    struct node * q;

    while( p != NULL )
    {
        q = p;
        p = p->next_node;
        free(q);
    }
}

// -----------------------------------------------------------------------
//                      QVFStudy::DeleteStrongFocus
// -----------------------------------------------------------------------

void QVFStudy::DeleteStrongFocus( struct strong_focus * p )
{
    struct strong_focus * q;

    while( p != NULL )
    {
        q = p;
        p = p->next_sf;
        free(q);
    }
}

// -----------------------------------------------------------------------
//                          QVFStudy::DeleteWeakFocus
// -----------------------------------------------------------------------

void QVFStudy::DeleteWeakFocus( struct weak_focus * p )
{
    struct weak_focus * q;

    while( p != NULL )
    {
        q = p;
        p = p->next_wf;
        free(q);
    }
}

// -----------------------------------------------------------------------
//                      QVFStudy::DeleteDegenerate
// -----------------------------------------------------------------------

void QVFStudy::DeleteDegenerate( struct degenerate * p )
{
    struct degenerate * q;

    while( p != NULL )
    {
        q = p;
        p = p->next_de;

        if( q->dummy )
            DeleteBlowup( q->blow_up );
        free(q);
    }
}

// -----------------------------------------------------------------------
//                      QVFStudy::DeleteSeparatrices
// -----------------------------------------------------------------------

void QVFStudy::DeleteSeparatrices( struct sep * p )
{
    struct sep * q;

    while( p != NULL )
    {
        q = p;
        p = p->next_sep;

        DeleteOrbitPoint( q->first_sep_point );
        if( q->dummy )
            delete_term1( q->separatrice );
        free( q );
    }
}
   
// -----------------------------------------------------------------------
//                      QVFStudy::DeleteTransformations
// -----------------------------------------------------------------------

void QVFStudy::DeleteTransformations( struct transformations * t )
{
    struct transformations * u;

    while( t != NULL )
    {
        u = t;
        t = t->next_trans;
        free(u);
    }
}

// -----------------------------------------------------------------------
//                      QVFStudy::DeleteBlowup
// -----------------------------------------------------------------------

void QVFStudy::DeleteBlowup( struct blow_up_points * b )
{
    struct blow_up_points * c;

    while( b != NULL )
    {
        c = b;
        b = b->next_blow_up_point;
        DeleteTransformations( c->trans );
        delete_term2( c->vector_field[0] );
        delete_term2( c->vector_field[1] );
        delete_term1( c->sep );
        DeleteOrbitPoint( c->first_sep_point );
        free(c);
    }
}

// -----------------------------------------------------------------------
//                      QVFStudy::DeleteLimitCycle
// -----------------------------------------------------------------------

void QVFStudy::DeleteLimitCycle( struct orbits * p )
{
    DeleteOrbit(p); // limit cycle is implemented as orbit.
}

// -----------------------------------------------------------------------
//                  QVFStudy::DeleteOrbitPoint
// -----------------------------------------------------------------------

void QVFStudy::DeleteOrbitPoint( P4ORBIT p )
{
    P4ORBIT q;
    
    while( p != NULL )
    {
        q = p;
        p = p->next_point;

        free(q);
    }
}

// -----------------------------------------------------------------------
//                      QVFStudy::DeleteOrbit
// -----------------------------------------------------------------------

void QVFStudy::DeleteOrbit( struct orbits * p )
{
    struct orbits * q;

    while( p !=NULL )
    {
        q = p;
        p = p->next_orbit;
        
        DeleteOrbitPoint( q->f_orbits );
        free(q);
   }
}

// -----------------------------------------------------------------------
//                      QVFStudy::ReadTables
// -----------------------------------------------------------------------
//
// read filename_vec.tab
// read filename_inf.tab
// read filename_fin.tab

bool QVFStudy::ReadTables( QString basename )
{
    FILE * fp;
    int j;

    DeleteVF();     // initialize structures, delete previous vector field if any

    fp = fopen( QFile::encodeName( basename + "_vec.tab" ), "rt" );
    if( fp == NULL )
    {
        DeleteVF();
        return false;
    }

    if( fscanf( fp, "%d %d %d ", &typeofstudy, &p, &q ) != 3 )
    {
        DeleteVF();
        fclose(fp);
        return false;
    }

    if( typeofstudy == TYPEOFSTUDY_ONE )
    {
        if( fscanf( fp, "%lf %lf %lf %lf", &xmin, &xmax, &ymin, &ymax ) != 4 )
        {
            DeleteVF();
            fclose(fp);
            return false;
        }
        p = q = 1;
        typeofview = TYPEOFVIEW_PLANE;
    }
    else
        typeofview = TYPEOFVIEW_SPHERE;

    plweights = (( p == 1 && q == 1 ) ? false : true);

    double_p = (double)p;
    double_q = (double)q;
    double_p_plus_q = (double)(p+q);
    double_p_minus_1 = (double)(p-1);
    double_q_minus_1 = (double)(q-1);
    double_q_minus_p = (double)(q-p);

    if( !ReadGCF( fp ) )
    {
        DeleteVF();
        fclose(fp);
        return false;
    }

    if( !ReadVectorField( fp, f_vec_field ) )
    {
        DeleteVF();
        fclose(fp);
        return false;
    }

    if( !ReadVectorField( fp, vec_field_U1 ) )
    {
        DeleteVF();
        fclose(fp);
        return false;
    }

    if( !ReadVectorField( fp, vec_field_V1 ) )
    {
        DeleteVF();
        fclose(fp);
        return false;
    }

    if( !ReadVectorField( fp, vec_field_U2 ) )
    {
        DeleteVF();
        fclose(fp);
        return false;
    }

    if( !ReadVectorField( fp, vec_field_V2 ) )
    {
        DeleteVF();
        fclose(fp);
        return false;
    }

    if( plweights )
    {
        if( !ReadVectorFieldCylinder( fp, vec_field_C ) )
        {
            DeleteVF();
            fclose(fp);
            return false;
        }
        singinf = 0;
    }
    else
    {
        if( fscanf( fp, "%d %d", &singinf, &VFResults.dir_vec_field ) != 2 )
        {
            DeleteVF();
            fclose(fp);
            return false;
        }
    }

    fclose( fp );

    if( typeofstudy != TYPEOFSTUDY_INF )
    {
        fp = fopen( QFile::encodeName( basename + "_fin.tab" ), "rt" );
        if( fp != NULL )
        {
            if( !ReadPoints( fp ) )
            {
                DeleteVF();
                fclose( fp );
                return false;
            }
            fclose( fp );
        }
        else
        {
            DeleteVF();
            return false;
        }
    }

    if( typeofstudy != TYPEOFSTUDY_ONE && typeofstudy != TYPEOFSTUDY_FIN )
    {
        fp = fopen( QFile::encodeName( basename + "_inf.tab" ), "rt" );
        if( fp != NULL )
        {
            if( p==1 && q==1 )
            {
                for( j = 1; j <= 2; j++ )
                {
                    if( !ReadPoints( fp ) )
                    {
                        DeleteVF();
                        fclose(fp);
                        return false;
                    }
                }
            }
            else
            {
                for( j = 1; j <= 4; j++ )
                {
                    if( !ReadPoints( fp ) )
                    {
                        DeleteVF();
                        fclose(fp);
                        return false;
                    }
                }
            }
            fclose(fp);
        }
        else
        {
            DeleteVF();
            return false;
        }
    }

    //Dump(basename);
    return true;
} 

// -----------------------------------------------------------------------
//                      QVFStudy::ReadGCF
// -----------------------------------------------------------------------

bool QVFStudy::ReadGCF( FILE * fp )
{
    int N, degree_gcf;

    if( fscanf( fp, "%d", &degree_gcf ) != 1 )
        return false;

    if( degree_gcf )
    {
        if( fscanf( fp, "%d", &N ) != 1 )
            return false;

        gcf = (P4POLYNOM2)malloc( sizeof(struct term2) );
        gcf->next_term2 = NULL;

        if( !ReadTerm2( fp, gcf, N ) )
            return false;

        if( fscanf( fp, "%d", &N ) != 1 )
            return false;

        gcf_U1= (P4POLYNOM2)malloc( sizeof(struct term2) );
        gcf_U1->next_term2 = NULL;

        if( !ReadTerm2( fp, gcf_U1, N ) )
            return false;

        if( fscanf( fp, "%d", &N ) != 1 )
            return false;

        gcf_U2= (P4POLYNOM2)malloc( sizeof(struct term2) );
        gcf_U2->next_term2 = NULL;

        if( !ReadTerm2( fp, gcf_U2, N ) )
            return false;

        if( fscanf( fp, "%d", &N ) != 1 )
            return false;

        gcf_V1= (P4POLYNOM2)malloc( sizeof(struct term2) );
        gcf_V1->next_term2 = NULL;
        if( !ReadTerm2( fp, gcf_V1, N ) )
            return false;
        
        if( fscanf( fp, "%d", &N ) != 1 )
            return false;
        gcf_V2= (P4POLYNOM2)malloc( sizeof(struct term2) );
        gcf_V2->next_term2 = NULL;
        if( !ReadTerm2( fp, gcf_V2, N ) )
            return false;

        if( p!=1 || q!=1 )
        {
            if( fscanf( fp, "%d", &N ) != 1 )
                return false;

            gcf_C = (P4POLYNOM3)malloc( sizeof(struct term3) );
            gcf_C->next_term3 = NULL;
            if( !ReadTerm3( fp, gcf_C, N ) )
                return false;
        }
    }
    else
    {
        gcf = NULL;
        gcf_U1 = NULL;
        gcf_U2 = NULL;
        gcf_V1 = NULL;
        gcf_V2 = NULL;
        gcf_C = NULL;
    }

    return true;
}

// -----------------------------------------------------------------------
//                      QVFStudy::ReadVectorField
// -----------------------------------------------------------------------

bool QVFStudy::ReadVectorField( FILE * fp, P4POLYNOM2 * vf )
{
    int M,N;

    vf[0] = (P4POLYNOM2)malloc( sizeof( struct term2 ) ); vf[0]->next_term2 = NULL;
    vf[1] = (P4POLYNOM2)malloc( sizeof( struct term2 ) ); vf[1]->next_term2 = NULL;

    if( fscanf( fp, "%d", &M ) != 1 )
        return false;
    if( !ReadTerm2( fp, vf[0], M ) )
        return false;
    if( fscanf( fp, "%d", &N ) != 1 )
        return false;
    if( !ReadTerm2( fp, vf[1], N ) )
        return false;

    return true;
}

// -----------------------------------------------------------------------
//                      QVFStudy::ReadVectorFieldCylinder
// -----------------------------------------------------------------------

bool QVFStudy::ReadVectorFieldCylinder( FILE * fp, P4POLYNOM3 * vf )
{
    int N;

    vf[0] = (P4POLYNOM3)malloc( sizeof( struct term3 ) ); vf[0]->next_term3 = NULL;
    vf[1] = (P4POLYNOM3)malloc( sizeof( struct term3 ) ); vf[1]->next_term3 = NULL;

    if( fscanf( fp, "%d", &N ) != 1 )
        return false;
    if( !ReadTerm3( fp, vf[0], N ) )
        return false;
    if( fscanf( fp, "%d", &N ) != 1 )
        return false;
    if( !ReadTerm3( fp, vf[1], N ) )
        return false;

    return true;
}

// -----------------------------------------------------------------------
//                      QVFStudy::ReadPoints
// -----------------------------------------------------------------------

bool QVFStudy::ReadPoints( FILE * fp )
{
    int N, i, typ;

    if( fscanf( fp, "%d ", &N ) != 1 )
        return false;

    for( i = 1; i <= N; i++ )
    {
        if( fscanf( fp, "%d ", &typ ) != 1 )
            return false;

        switch( typ )
        {
        case SADDLE:
            if( !ReadSaddlePoint( fp ) )
                return false;
            break;
        case SEMI_HYPERBOLIC:
            if( !ReadSemiElementaryPoint( fp ) )
                return false;
            break;
        case NODE:
            if( !ReadNodePoint( fp ) )
                return false;
            break;
        case STRONG_FOCUS:
            if( !ReadStrongFocusPoint( fp ) )
                return false;
            break;
        case WEAK_FOCUS:
            if( !ReadWeakFocusPoint( fp ) )
                return false;
            break;
        case NON_ELEMENTARY:
            if( !ReadDegeneratePoint( fp ) )
                return false;
            break;
        default:
            return false;
        }
    }

    return true;
}

// -----------------------------------------------------------------------
//                      QVFStudy::ReadTerm1
// -----------------------------------------------------------------------

bool QVFStudy::ReadTerm1( FILE * fp, P4POLYNOM1 p, int N )
{
    int i;
    P4POLYNOM1 _p;

    _p = p;

    p->next_term1 = NULL;
    if( fscanf( fp, "%d %lf", &(p->exp), &(p->coeff) ) != 2 )
        return false;

    for( i = 2; i <= N; i++ )
    {
        p->next_term1 = (P4POLYNOM1)malloc( sizeof( struct term1 ) );
        p = p->next_term1;
        p->next_term1 = NULL;
        if( fscanf( fp, "%d %lf", &(p->exp), &(p->coeff) ) != 2 )
        {
            delete_term1( _p->next_term1 );
            _p->next_term1 = NULL;
            return false;
        }
    }
    return true;
}

// -----------------------------------------------------------------------
//                      QVFStudy::ReadTerm2
// -----------------------------------------------------------------------

bool QVFStudy::ReadTerm2( FILE * fp, P4POLYNOM2 p, int N )
{
    int i;
    P4POLYNOM2 _p;

    _p = p;
    p->next_term2 = NULL;
    if( fscanf( fp, "%d %d %lf", &(p->exp_x), &(p->exp_y), &(p->coeff) ) != 3 )
        return false;

    for( i = 2; i <= N; i++ )
    {
        p->next_term2 = (P4POLYNOM2)malloc( sizeof( struct term2 ) );
        p = p->next_term2;
        p->next_term2 = NULL;
        if( fscanf( fp, "%d %d %lf", &(p->exp_x), &(p->exp_y), &(p->coeff) ) != 3 )
        {
            delete_term2( _p->next_term2 );
            _p->next_term2 = NULL;
            return false;
        }
    }
    return true;
}

// -----------------------------------------------------------------------
//                      QVFStudy::ReadTerm3
// -----------------------------------------------------------------------

bool QVFStudy::ReadTerm3( FILE * fp, P4POLYNOM3 p, int N )
{
    int i;
    P4POLYNOM3 _p;
    _p = p;
    p->next_term3 = NULL;
    if( fscanf( fp, "%d %d %d %lf", &(p->exp_r), &(p->exp_Co), &(p->exp_Si), &(p->coeff) ) != 4 )
        return false;

    for( i = 2; i <= N; i++ )
    {
        p->next_term3 = (P4POLYNOM3)malloc( sizeof( struct term3 ) );
        p = p->next_term3;
        p->next_term3 = NULL;
        if( fscanf( fp, "%d %d %d %lf", &(p->exp_r), &(p->exp_Co), &(p->exp_Si), &(p->coeff) ) != 4 )
        {
            delete_term3( _p->next_term3 );
            _p->next_term3 = NULL;
            return false;
        }
    }
    return true;
}

// -----------------------------------------------------------------------
//                      QVFStudy::ReadSaddlePoint
// -----------------------------------------------------------------------

bool QVFStudy::ReadSaddlePoint( FILE * fp )
{
    int N;
    struct sep * sep1;
    struct sep * sep2;

    // make room in structure

    struct saddle * last;
    struct saddle * point;

    last = NULL;
    point = first_saddle_point;
    while( point != NULL )
    {
        last = point;
        point = point->next_saddle;
    }

    point = (struct saddle *)malloc( sizeof(struct saddle) );
    if( last == NULL )
        first_saddle_point = point;
    else
        last->next_saddle = point;

    point->next_saddle = NULL;

    // fill structure

    fscanf( fp, "%lf %lf", &(point->x0), &(point->y0) );
    fscanf( fp, "%lf %lf %lf %lf", &(point->a11), &(point->a12), &(point->a21), &(point->a22) );

    ReadVectorField( fp, point->vector_field );
    fscanf( fp, "%d ", &(point->chart) );
    point->separatrices = (struct sep *)malloc( sizeof(struct sep) );
    sep1 = point->separatrices;
    fscanf( fp, "%d ", &(sep1->type) );
    fscanf( fp, "%d ", &N );

    sep1->dummy = 1;
    sep1->separatrice = (P4POLYNOM1)malloc( sizeof(struct term1) );
    ReadTerm1( fp, sep1->separatrice, N );
    sep1->direction = 1;
    sep1->d = 0;
    sep1->first_sep_point = NULL;
    sep1->last_sep_point = NULL;
    sep1->next_sep = NULL;

    if( point->chart == CHART_R2 || singinf )
    {
        // point is finite hence we have 4 separatrices or we have a line of singularities
        // at infinity and hence we have also 4 separatrices if we delete the line

        sep2 = (struct sep *)malloc( sizeof(struct sep) );
        sep1->next_sep = sep2;
        
        sep2->type = sep1->type;
        sep2->direction = -1;
        sep2->d = 0;
        sep2->dummy = 0;
        sep2->separatrice = sep1->separatrice;
        sep2->first_sep_point = NULL;
        sep2->last_sep_point = NULL;

        sep1 = sep2->next_sep = (struct sep *)malloc( sizeof(struct sep) );

        fscanf( fp, "%d", &(sep1->type) );
        fscanf( fp, "%d ", &N );

        sep1->dummy = 1;
        sep1->separatrice = (P4POLYNOM1)malloc( sizeof(struct term1) );
        ReadTerm1( fp, sep1->separatrice, N );
        sep1->direction = 1;
        sep1->d = 1;
        sep1->first_sep_point = NULL;
        sep1->last_sep_point = NULL;

        sep2 = (struct sep *)malloc( sizeof(struct sep) );
        sep1->next_sep = sep2;

        sep2->type = sep1->type;
        sep2->direction = -1;
        sep2->d = 1;
        sep2->dummy = 0;
        sep2->separatrice = sep1->separatrice;
        sep2->first_sep_point = NULL;
        sep2->last_sep_point = NULL;
        sep2->next_sep = NULL;
    }

    fscanf( fp, "%lf ", &(point->epsilon) );
    point->dummy = 1;
 
    // line at infinity a line of singularities in poincare disc
 
    if( singinf && point->chart != CHART_R2 )
    {
        last = point;
        point = (struct saddle *)malloc( sizeof(struct saddle) );
        last->next_saddle = point;
        point->next_saddle = NULL;

        point->x0 = last->x0;
        point->y0 = 0.0;
        point->chart = ((last->chart == CHART_U1) ? CHART_V1 : CHART_V2);
        point->dummy = 0;
        point->separatrices = last->separatrices;
        point->epsilon = last->epsilon;
        point->a11 = last->a11;
        point->a12 = last->a12;
        point->a21 = last->a21;
        point->a22 = last->a22;
        point->vector_field[0] = NULL;
        point->vector_field[1] = NULL;
    } 

    return true;
}

// -----------------------------------------------------------------------
//                      QVFStudy::ReadSemiElementaryPoint
// -----------------------------------------------------------------------

bool QVFStudy::ReadSemiElementaryPoint( FILE * fp )
{
    // make room in structure

    struct semi_elementary * last;
    struct semi_elementary * point;
    int N;
    int s;
    struct sep * sep1;
    double y[2];
    bool ok;

    last = NULL;
    point = first_se_point;
    while( point != NULL )
    {
        last = point;
        point = point->next_se;
    }

    point = (struct semi_elementary *)malloc( sizeof(struct semi_elementary) );
    if( last == NULL )
        first_se_point = point;
    else
        last->next_se = point;

    point->next_se = NULL;

    fscanf( fp, "%lf %lf ", &(point->x0), &(point->y0) );
    fscanf( fp, "%lf %lf %lf %lf ", &(point->a11), &(point->a12),
                                    &(point->a21), &(point->a22) );
    ReadVectorField( fp, point->vector_field );
    fscanf( fp, "%d %d %d", &(point->type), &s, &(point->chart) );

    switch( point->type )
    {
    case 1: // saddle-node 
        if( s && (point->chart != CHART_R2) )
            point->separatrices = NULL;
        else
        {
            point->separatrices = (struct sep *)malloc( sizeof(struct sep) );
            sep1 = point->separatrices;
            
            if( s )
                sep1->type = OT_CENT_STABLE;
            else
                sep1->type = OT_UNSTABLE;

            sep1->d = 0;
            if(((p==1) && (q==1) && ((point->chart == CHART_V1) || (point->chart==CHART_V2))) ||
                (point->chart==CHART_R2) )
            {
                sep1->direction = -1;
            }
            else
            {
                sep1->direction = 1;
            }

            fscanf( fp, "%d ", &N );
            sep1->dummy = 1;
            sep1->separatrice = (struct term1 *)malloc( sizeof(struct term1) );
            ReadTerm1( fp, sep1->separatrice, N );
            sep1->first_sep_point = NULL;
            sep1->last_sep_point = NULL;
            sep1->next_sep = NULL;
            if( point->chart == CHART_R2 )
            {
                sep1->next_sep = (struct sep *)malloc( sizeof(struct sep) );
                sep1 = sep1->next_sep;
                sep1->type = OT_UNSTABLE;
                sep1->d = 1;
                sep1->direction = 1;
                fscanf( fp, "%d ", &N );
                sep1->dummy = 1;
                sep1->separatrice = (struct term1 *)malloc( sizeof(struct term1) );
                ReadTerm1( fp, sep1->separatrice, N );
                sep1->first_sep_point = NULL;
                sep1->last_sep_point = NULL;
                sep1->next_sep = (struct sep *)malloc( sizeof(struct sep) );
                sep1->next_sep->type = OT_UNSTABLE;
                sep1->next_sep->d = 1;
                sep1->next_sep->direction = -1;
                sep1->next_sep->dummy = 0;
                sep1->next_sep->separatrice = sep1->separatrice;
                sep1->next_sep->first_sep_point = NULL;
                sep1->next_sep->last_sep_point = NULL;
                sep1->next_sep->next_sep = NULL;
            }
        }
        break;


    case 2: // saddle-node 
        point->separatrices = (struct sep *) malloc(sizeof(struct sep));
        sep1 = point->separatrices;
        if( s )
            sep1->type = STYPE_CENUNSTABLE;
        else
            sep1->type = STYPE_STABLE;

        sep1->d = 0; 
        if( (p==1) && (q==1) && ((point->chart==CHART_V1) || (point->chart==CHART_V2)) )
            sep1->direction = -1;
        else
            sep1->direction = 1;

        fscanf( fp, "%d ", &N );
        sep1->dummy = 1;
        sep1->separatrice = (struct term1 *) malloc(sizeof(struct term1));
        ReadTerm1( fp, sep1->separatrice, N );
        sep1->first_sep_point = NULL;
        sep1->last_sep_point = NULL;
        sep1->next_sep = NULL;
        if( point->chart == CHART_R2 )
        {
            sep1->next_sep = (struct sep *) malloc( sizeof(struct sep) );
            sep1 = sep1->next_sep;
            sep1->type = STYPE_STABLE;
            sep1->d = 1;
            sep1->direction = 1;
            fscanf( fp, "%d ", &N );
            sep1->dummy = 1;
            sep1->separatrice = (struct term1 *)malloc( sizeof(struct term1) );
            ReadTerm1( fp, sep1->separatrice, N );
            sep1->first_sep_point = NULL;
            sep1->last_sep_point = NULL;
            sep1->next_sep = (struct sep *)malloc(sizeof(struct sep));
            sep1->next_sep->type = STYPE_STABLE;
            sep1->next_sep->d = 1;
            sep1->next_sep->direction = -1;
            sep1->next_sep->dummy = 0;
            sep1->next_sep->separatrice = sep1->separatrice;
            sep1->next_sep->first_sep_point = NULL;
            sep1->next_sep->last_sep_point = NULL;
            sep1->next_sep->next_sep = NULL;
        }
        break;

    case 3: // saddle-node 
        point->separatrices = (struct sep *) malloc(sizeof(struct sep));
        sep1 = point->separatrices;
        if( s )
            sep1->type = STYPE_CENSTABLE;
        else
            sep1->type = STYPE_UNSTABLE;
        sep1->d = 0;
        if( (p==1) && (q==1) && ((point->chart==CHART_V1) || (point->chart==CHART_V2)) )
            sep1->direction = -1;
        else
            sep1->direction = 1;

        fscanf( fp, "%d ", &N );
        sep1->dummy = 1;
        sep1->separatrice = (struct term1 *) malloc(sizeof(struct term1));
        ReadTerm1( fp, sep1->separatrice, N );
        sep1->first_sep_point = NULL;
        sep1->last_sep_point = NULL;
        sep1->next_sep = NULL;
        if( point->chart == CHART_R2 )
        {
            sep1->next_sep = (struct sep *) malloc(sizeof(struct sep));
            sep1 = sep1->next_sep;
            sep1->type = STYPE_UNSTABLE; sep1->d = 1;
            sep1->direction = 1;
            fscanf( fp, "%d ", &N );
            sep1->dummy = 1;
            sep1->separatrice = (struct term1 *) malloc(sizeof(struct term1));
            ReadTerm1( fp, sep1->separatrice, N );
            sep1->first_sep_point = NULL;
            sep1->last_sep_point = NULL;
            sep1->next_sep = (struct sep *) malloc(sizeof(struct sep));
            sep1->next_sep->type = STYPE_UNSTABLE;
            sep1->next_sep->d = 1;
            sep1->next_sep->direction = -1;
            sep1->next_sep->dummy = 0;
            sep1->next_sep->separatrice = sep1->separatrice;
            sep1->next_sep->first_sep_point = NULL;
            sep1->next_sep->last_sep_point = NULL;
            sep1->next_sep->next_sep = NULL;
        }
        break;   

    case 4: // saddle-node 
        if( s && (point->chart != CHART_R2) )
        {
            point->separatrices = NULL;
        }
        else
        {
            point->separatrices = (struct sep *) malloc(sizeof(struct sep));
            sep1 = point->separatrices;
            if( s )
                sep1->type = STYPE_CENUNSTABLE;
            else
                sep1->type = STYPE_STABLE;

            sep1->d = 0; 
            if(((p == 1) && (q == 1) && ((point->chart == CHART_V1) || (point->chart == CHART_V2))) ||
                (point->chart == CHART_R2))
                sep1->direction = -1;
            else
                sep1->direction = 1;

            fscanf( fp, "%d ", &N );
            sep1->dummy = 1;
            sep1->separatrice = (struct term1 *) malloc(sizeof(struct term1));
            ReadTerm1( fp, sep1->separatrice, N );
            sep1->first_sep_point = NULL;
            sep1->last_sep_point = NULL;
            sep1->next_sep = NULL;
            if( point->chart == CHART_R2 )
            {
                sep1->next_sep = (struct sep *) malloc(sizeof(struct sep));
                sep1 = sep1->next_sep;
                sep1->type = STYPE_STABLE;
                sep1->d = 1;
                sep1->direction = 1;
                fscanf( fp, "%d ", &N );
                sep1->dummy = 1;
                sep1->separatrice = (struct term1 *) malloc(sizeof(struct term1));
                ReadTerm1( fp, sep1->separatrice, N );
                sep1->first_sep_point = NULL;
                sep1->last_sep_point = NULL;
                sep1->next_sep = (struct sep *) malloc(sizeof(struct sep));
                sep1->next_sep->type = STYPE_STABLE;
                sep1->next_sep->d = 1;
                sep1->next_sep->direction = -1;
                sep1->next_sep->dummy = 0;
                sep1->next_sep->separatrice = sep1->separatrice;
                sep1->next_sep->first_sep_point = NULL;
                sep1->next_sep->last_sep_point = NULL;
                sep1->next_sep->next_sep = NULL;
            }
        } 
        break;   

    case 6: // saddle 
        point->separatrices = (struct sep *) malloc(sizeof(struct sep));
        sep1 = point->separatrices;
        if( s )
            sep1->type = STYPE_CENUNSTABLE;
        else
            sep1->type = STYPE_STABLE;
        sep1->d = 0; 
        if( (p == 1) && (q == 1) && ((point->chart== CHART_V1) || (point->chart== CHART_V2)) )
            sep1->direction = -1;
        else
            sep1->direction = 1;
        fscanf( fp, "%d ", &N );
        sep1->dummy = 1;
        sep1->separatrice = (struct term1 *) malloc(sizeof(struct term1));
        ReadTerm1( fp, sep1->separatrice, N );
        sep1->first_sep_point = NULL;
        sep1->last_sep_point = NULL;
        sep1->next_sep = NULL;
        if( point->chart== CHART_R2 )
        {
            sep1->next_sep = (struct sep *) malloc(sizeof(struct sep));
            sep1->next_sep->type = STYPE_CENUNSTABLE;
            sep1->next_sep->d = 0;
            sep1->next_sep->direction = -1;
            sep1->next_sep->dummy = 0;
            sep1->next_sep->separatrice = sep1->separatrice;
            sep1->next_sep->first_sep_point = NULL;
            sep1->next_sep->last_sep_point = NULL;
            sep1->next_sep->next_sep = (struct sep *) malloc(sizeof(struct sep));
            sep1 = sep1->next_sep->next_sep;
            sep1->type = STYPE_STABLE;
            sep1->d = 1;
            sep1->direction = 1;
            fscanf( fp, "%d ", &N );
            sep1->dummy = 1;
            sep1->separatrice = (struct term1 *) malloc(sizeof(struct term1));
            ReadTerm1( fp, sep1->separatrice, N);
            sep1->first_sep_point = NULL;
            sep1->last_sep_point = NULL;
            sep1->next_sep = (struct sep *) malloc(sizeof(struct sep));
            sep1->next_sep->type = STYPE_STABLE;
            sep1->next_sep->d = 1;
            sep1->next_sep->direction = -1;
            sep1->next_sep->dummy = 0;
            sep1->next_sep->separatrice = sep1->separatrice;
            sep1->next_sep->first_sep_point = NULL;
            sep1->next_sep->last_sep_point = NULL;
            sep1->next_sep->next_sep = NULL;
        }
        break;

    case 7: // saddle 
        point->separatrices = (struct sep *) malloc(sizeof(struct sep));
        sep1 = point->separatrices;
        if( s )
            sep1->type = STYPE_CENSTABLE;
        else
            sep1->type = STYPE_UNSTABLE;
        sep1->d = 0; 
        if( (p==1) && (q==1) && ((point->chart==CHART_V1) || (point->chart==CHART_V2)) )
            sep1->direction = -1;
        else
            sep1->direction = 1;
        fscanf( fp, "%d ", &N );
        sep1->dummy = 1;
        sep1->separatrice = (struct term1 *) malloc(sizeof(struct term1));
        ReadTerm1( fp, sep1->separatrice, N);
        sep1->first_sep_point = NULL;
        sep1->last_sep_point = NULL;
        sep1->next_sep = NULL;
        if( point->chart== CHART_R2 )
        {
            sep1->next_sep = (struct sep *) malloc(sizeof(struct sep));
            sep1->next_sep->type = STYPE_CENSTABLE;
            sep1->next_sep->d = 0;
            sep1->next_sep->direction = -1;
            sep1->next_sep->dummy = 0;
            sep1->next_sep->separatrice = sep1->separatrice;
            sep1->next_sep->first_sep_point = NULL;
            sep1->next_sep->last_sep_point = NULL;
            sep1->next_sep->next_sep = (struct sep *) malloc(sizeof(struct sep));
            sep1 = sep1->next_sep->next_sep;
            sep1->type = STYPE_UNSTABLE;
            sep1->d = 1;
            sep1->direction = 1;
            fscanf( fp, "%d ",&N );
            sep1->dummy = 1;
            sep1->separatrice = (struct term1 *) malloc(sizeof(struct term1));
            ReadTerm1( fp, sep1->separatrice, N );
            sep1->first_sep_point = NULL;
            sep1->last_sep_point = NULL;
            sep1->next_sep = (struct sep *) malloc(sizeof(struct sep));
            sep1->next_sep->type = STYPE_UNSTABLE;
            sep1->next_sep->d = 1;
            sep1->next_sep->direction = -1;
            sep1->next_sep->dummy = 0;
            sep1->next_sep->separatrice = sep1->separatrice;
            sep1->next_sep->first_sep_point = NULL;
            sep1->next_sep->last_sep_point = NULL;
            sep1->next_sep->next_sep = NULL;
        }
        break;

    default:
        point->separatrices = NULL; 

        // change type of node if we have a gcf ? 
        y[0] = point->x0;
        y[1] = point->y0;
        ok = false;
        switch( point->chart )
        {
        case CHART_R2:
            if( eval_term2( gcf, y ) < 0 )
                ok = true;
            break;
        case CHART_U1:
            if( eval_term2( gcf_U1, y ) < 0 )
                ok = true;
            break;
        case CHART_V1:
            if((p==1) && (q==1))
                y[0] = -y[0];
            if( eval_term2( gcf_V1, y ) < 0 )
                ok = true;
            break;        
        case CHART_U2:
            if( eval_term2( gcf_U2, y ) < 0 )
                ok = true;
            break;
        case CHART_V2:
            if((p==1) && (q==1))
                y[0] = -y[0];
            if( eval_term2( gcf_V2, y ) < 0 )
                ok = true;
            break;  
        }
        if( ok )
            if( point->type == 5 )
                point->type = 8;
            else
                point->type = 5;
    }

    fscanf( fp, "%lf ", &(point->epsilon) );
    point->dummy = 1;

    // line at infinity a line of singularities in poincare disc 

    if( singinf && (point->chart != CHART_R2) )
    {
        point->next_se = (struct semi_elementary *) malloc(sizeof(struct semi_elementary));
        point->next_se->x0 = point->x0;
        point->next_se->y0 = 0.0;
        if( VFResults.dir_vec_field == 1 )
            point->next_se->type = point->type;
        else
            switch( point->type )
            {
            case 1: point->next_se->type = 3; break;
            case 2: point->next_se->type = 4; break;
            case 3: point->next_se->type = 1; break;
            case 4: point->next_se->type = 2; break;
            case 6: point->next_se->type = 7; break;
            case 7: point->next_se->type = 6; break;
            default:point->next_se->type = point->type;
            }

        point->next_se->chart = (point->chart== CHART_U1) ? CHART_V1: CHART_V2;
        point->next_se->dummy = 0;
        point->next_se->separatrices = point->separatrices; 
        point->next_se->epsilon = point->epsilon;
        point->next_se->a11 = point->a11;
        point->next_se->a12 = point->a12;
        point->next_se->a21 = point->a21;
        point->next_se->a22 = point->a22;
        point->next_se->vector_field[0] = NULL;
        point->next_se->vector_field[1] = NULL;
        point->next_se->next_se = NULL;
    } 
    return true;
}

// -----------------------------------------------------------------------
//                      QVFStudy::ReadStrongFocusPoint
// -----------------------------------------------------------------------

bool QVFStudy::ReadStrongFocusPoint( FILE * fp )
{
    double y[2];

    // make room in structure

    struct strong_focus * last;
    struct strong_focus * point;

    last = NULL;
    point = first_sf_point;
    while( point != NULL )
    {
        last = point;
        point = point->next_sf;
    }

    point = (struct strong_focus *)malloc( sizeof(struct strong_focus) );
    if( last == NULL )
        first_sf_point = point;
    else
        last->next_sf = point;

    point->next_sf = NULL;

    // fill structure

    fscanf( fp, "%d %lf %lf ", &(point->stable), &(point->x0), &(point->y0) );
    fscanf( fp, "%d ", &(point->chart) );

    // change type of node if we have a gcf ? 

    y[0] = point->x0;
    y[1] = point->y0;

    switch( point->chart )
    {
    case CHART_R2:
        if( eval_term2( gcf, y ) < 0 )
            point->stable *= -1;
        break;

    case CHART_U1:
        if( eval_term2( gcf_U1, y ) < 0 )
            point->stable *= -1;
        break;

    case CHART_V1:
        if( p==1 && q==1 )
            y[0] =- y[0];

        if( eval_term2( gcf_V1, y ) < 0 )
            point->stable *= -1;
        break;

    case CHART_U2:
        if( eval_term2( gcf_U2, y ) < 0 )
            point->stable *= -1;
        break;

    case CHART_V2:
        if( p==1 && q==1 )
            y[0] = -y[0];

        if( eval_term2( gcf_V2, y ) < 0 )
            point->stable *= -1;
        break;
    }

    // line at infinity a line of singularities in poincare disc
    
    if( singinf && point->chart != CHART_R2 )
    {
        last = point;
        point = (struct strong_focus *)malloc( sizeof(struct strong_focus) );
        last->next_sf = point;
        point->next_sf = NULL;

        point->x0 = last->x0;
        point->y0 = 0.0;
        point->chart = (( point->chart == CHART_U1 ) ? CHART_V1 : CHART_V2 );
        point->stable = last->stable * ((VFResults.dir_vec_field == -1) ? -1 : 1);
    }  

    return true;
}

bool QVFStudy::ReadWeakFocusPoint( FILE * fp )
{
    // make room in structure

    struct weak_focus * last;
    struct weak_focus * point;
    double y[2];

    last = NULL;
    point = first_wf_point;
    while( point != NULL )
    {
        last = point;
        point = point->next_wf;
    }

    point = (struct weak_focus *)malloc( sizeof(struct weak_focus) );
    if( last == NULL )
        first_wf_point = point;
    else
        last->next_wf = point;

    point->next_wf = NULL;

    fscanf( fp, "%lf %lf ", &(point->x0), &(point->y0) );
    fscanf( fp, "%d %d ", &(point->type), &(point->chart) ); 

    ///* change type of node if we have a gcf ? 
 
    if( ( point->type == STYPE_STABLE ) || ( point->type == STYPE_UNSTABLE) )
    {
        y[0] = point->x0;
        y[1] = point->y0;

        switch( point->chart )
        {
        case CHART_R2:
            if( eval_term2( gcf, y ) < 0 )
                point->type *= -1;
            break;

        case CHART_U1:
            if( eval_term2( gcf_U1, y ) < 0 )
                point->type *= -1;
            break;

        case CHART_V1:
            if( (p==1) && (q==1) )
                y[0] = -y[0];
            if( eval_term2( gcf_V1, y ) < 0 )
                point->type *= -1;
            break;

        case CHART_U2:
            if( eval_term2( gcf_U2, y ) < 0 )
                point->type *= -1;
            break;

        case CHART_V2:
            if( (p==1) && (q==1) )
                y[0] = -y[0];
            if( eval_term2( gcf_V2, y ) < 0 )
                point->type*=-1;
            break;
        } 
    }

    if( singinf && (point->chart != CHART_R2) )
    {
        point->next_wf = (struct weak_focus *) malloc(sizeof(struct weak_focus));
        point->next_wf->x0 = point->x0;
        point->next_wf->y0 = 0.0;

        if( VFResults.dir_vec_field == 1 )
            point->next_wf->type = point->type;
        else
            switch( point->type )
            {
                case STYPE_STABLE:      point->next_wf->type = STYPE_UNSTABLE;  break;
                case STYPE_UNSTABLE:    point->next_wf->type = STYPE_STABLE;    break;
                default:                point->next_wf->type = point->type;     break;
            }

        point->next_wf->chart = (point->chart==CHART_U1) ? CHART_V1: CHART_V2;
        point = point->next_wf;
        point->next_wf = NULL;
    }

    return true;
}

// -----------------------------------------------------------------------
//                  QVFStudy::ReadDegeneratePoint
// -----------------------------------------------------------------------

bool QVFStudy::ReadDegeneratePoint( FILE * fp )
{
    int n;

    // make room in structure

    struct degenerate * last;
    struct degenerate * point;

    last = NULL;
    point = first_de_point;
    while( point != NULL )
    {
        last = point;
        point = point->next_de;
    }

    point = (struct degenerate *)malloc( sizeof(struct degenerate) );
    if( last == NULL )
        first_de_point = point;
    else
        last->next_de = point;

    point->next_de = NULL;

    // load structure

    fscanf( fp, "%lf %lf %lf %d ", &(point->x0), &(point->y0), &(point->epsilon), &n );
    point->blow_up=NULL;
    if( n )
    {
        point->blow_up = (struct blow_up_points *)malloc( sizeof(struct blow_up_points) );
        ReadBlowupPoints( fp, point->blow_up, n );
        point->blow_up->blow_up_vec_field = true;
    }

    fscanf( fp, "%d ", &(point->chart) );
    
    point->dummy = 1;

    // line at infinity a line of singularities in poincare disc 

    if( singinf && point->chart != CHART_R2 )
    {
        last = point;
        point = (struct degenerate *)malloc( sizeof(struct degenerate) );
        last->next_de = point;
        point->next_de = NULL;

        point->x0 = last->x0;
        point->y0 = 0.0;
        point->chart = (last->chart== CHART_U1) ? CHART_V1 : CHART_V2;
        point->epsilon = last->epsilon;
        point->dummy = 0;
        point->blow_up = last->blow_up;
    }

    return true;
}

// -----------------------------------------------------------------------
//                      QVFStudy::ReadNodePoint
// -----------------------------------------------------------------------

bool QVFStudy::ReadNodePoint( FILE * fp )
{
    double y[2];

    // make room in structure

    struct node * last;
    struct node * point;

    last = NULL;
    point = first_node_point;
    while( point != NULL )
    {
        last = point;
        point = point->next_node;
    }

    point = (struct node *)malloc( sizeof(struct node) );
    if( last == NULL )
        first_node_point = point;
    else
        last->next_node = point;

    point->next_node = NULL;

    // load point structure

    fscanf( fp, "%lf %lf %d ", &(point->x0), &(point->y0), &(point->stable) );
    fscanf( fp, "%d ", &(point->chart) );

    // change type of node if we have a gcf ?

    y[0] = point->x0;
    y[1]= point->y0;
    
    switch( point->chart )
    {
    case CHART_R2:
        if( eval_term2(gcf,y) < 0 )
            point->stable *= -1;
        break;

    case CHART_U1:
        if( eval_term2(gcf_U1,y) < 0 )
            point->stable *= -1;
        break;

    case CHART_V1:
        if( p==1 && q==1 )
            y[0] = -y[0];

        if( eval_term2( gcf_V1, y ) < 0 )
            point->stable *= -1;
        break;

    case CHART_U2:
        if( eval_term2( gcf_U2, y ) < 0 )
            point->stable *= -1;
        break;

    case CHART_V2:
        if( p==1 && q==1)
            y[0] = -y[0];
        if( eval_term2( gcf_V2, y ) < 0 )
            point->stable *= -1;
        break;
    }
 
    // if line at infinity a line of singularities in poincare disc:
    // duplicate singularity by using a symmetry

    if( singinf != 0 && point->chart != CHART_R2 )
    {
        last = point;
        point = (struct node *)malloc( sizeof(struct node) );
        last->next_node = point;
        point->next_node = NULL;

        point->x0 = last->x0;
        point->y0 = 0.0;
        point->chart = (last->chart == CHART_U1) ? CHART_V1 : CHART_V2;
        point->stable = last->stable * ( VFResults.dir_vec_field == -1 ) ? -1 : 1;
    }

    return true;
}

// -----------------------------------------------------------------------
//                      QVFStudy::ReadTransformations
// -----------------------------------------------------------------------

bool QVFStudy::ReadTransformations( FILE * fp, struct transformations * trans, int n )
{

    int i;

    fscanf( fp, "%lf %lf %d %d %d %d %d %d %d",
                &(trans->x0), &(trans->y0), &(trans->c1), &(trans->c2),
                &(trans->d1),& (trans->d2), &(trans->d3), &(trans->d4), &(trans->d) );
    
    trans->next_trans = NULL;

    for( i = 2; i <= n; i++ )
    {
        trans->next_trans = (struct transformations *)malloc( sizeof(struct transformations) );
        trans = trans->next_trans;
    
        fscanf( fp, "%lf %lf %d %d %d %d %d %d %d",
                    &(trans->x0), &(trans->y0), &(trans->c1), &(trans->c2),
                    &(trans->d1),& (trans->d2), &(trans->d3), &(trans->d4), &(trans->d) );

        trans->next_trans=NULL;
    }

    return true;
}

// -----------------------------------------------------------------------
//                      QVFStudy::ReadBlowupPoints
// -----------------------------------------------------------------------

bool QVFStudy::ReadBlowupPoints( FILE * fp, struct blow_up_points * b, int n )
{
    int i, N, typ;

    for( i = 1; i <= n; i++ )
    {
        fscanf( fp, "%d ", &(b->n) );
        
        b->trans = (struct transformations *)malloc( sizeof(struct transformations) );

        ReadTransformations( fp, b->trans, b->n );
        fscanf( fp, "%lf %lf ", &(b->x0), &(b->y0) );
        fscanf( fp, "%lf %lf %lf %lf ",&(b->a11), &(b->a12), &(b->a21),&(b->a22) );
        ReadVectorField( fp, b->vector_field );
        b->sep = (struct term1 *)malloc( sizeof(struct term1) );
        fscanf( fp, "%d ", &N );
        ReadTerm1( fp, b->sep, N );
   
        fscanf( fp, "%d ", &typ );
        switch(typ )
        {
            case 1:  b->type = STYPE_UNSTABLE;          break;
            case 2:  b->type = STYPE_STABLE;            break;
            case 3:  b->type = STYPE_UNSTABLE;          break;
            case 4:  b->type = STYPE_STABLE;            break;
            case 5:  b->type = STYPE_CENUNSTABLE;       break;
            case 6:  b->type = STYPE_CENSTABLE;         break;
            case 7:  b->type = STYPE_CENUNSTABLE;       break; 
            case 8:  b->type = STYPE_CENSTABLE;         break;
            case 9:  b->type = STYPE_UNSTABLE;          break;
            case 10: b->type = STYPE_UNSTABLE;          break;
            case 11: b->type = STYPE_STABLE;            break;
            case 12: b->type = STYPE_STABLE;            break;                                                                     
        }
        b->first_sep_point = NULL;
        b->last_sep_point = NULL;
        b->next_blow_up_point = NULL;
        
        if( i < n )
        {
            b->next_blow_up_point = (struct blow_up_points *)malloc( sizeof(struct blow_up_points) );
            b = b->next_blow_up_point;
        }
    }
    return true;
}

#define DUMP(x) m->append( s.sprintf x );
#define DUMPSTR(x) m->append( x );

void DumpSeparatrices( QTextEdit * m, struct sep * separ, int margin )
{
    QString s;
    char smargin[80];
    char * stype;
    strcpy(smargin,"                              ");
    smargin[margin]=0;  // make number of spaces

    if( separ == NULL )
    {
        DUMP(( "%s/", smargin ))
            return;
    }

    while( separ != NULL )
    {
        switch( separ->type )
        {
        case STYPE_STABLE:      stype = "Stable         ";  break;
        case STYPE_UNSTABLE:    stype = "unstable       ";  break;
        case STYPE_CENSTABLE:   stype = "center-stable  ";  break;
        case STYPE_CENUNSTABLE: stype = "center-unstable";  break;
        default:                stype = "???????????????";  break;
        }

        DUMP(( "%sType=%s  Dir=%-2d  original=%d", smargin, stype, separ->direction, separ->dummy ))
        if( separ->d == 0 )
        {
            DUMP(( "%sTaylor: (x,y)=(t,%s)", smargin, DumpPoly1( separ->separatrice, "t" ) ))
        }
        else
        {
            DUMP(( "%sTaylor: (x,y)=(%s,t)", smargin, DumpPoly1( separ->separatrice, "t" ) ))
        }

        separ = separ->next_sep;
    }

}

void DumpSingularities( QTextEdit * m, struct genericsingularity * p, char * type, bool longversion )
{
    const char * chart;
    QString s;
    QByteArray ss;

    struct saddle * sa;
    struct degenerate * de;
    struct strong_focus * sf;
    struct weak_focus * wf;
    struct node * no;
    struct semi_elementary * se;

    while( p != NULL )
    {
        switch( p->chart )
        {
        case CHART_R2:  chart = "R2";   break;
        case CHART_U1:  chart = "U1";   break;
        case CHART_U2:  chart = "U2";   break;
        case CHART_V1:  chart = "V1";   break;
        case CHART_V2:  chart = "V2";   break;
        default:
            chart = "?";    break;
        }
        DUMP(( "%s (x0,y0)=(%g,%g)  Chart %s", type, (float)(p->x0), (float)(p->y0), chart ))
        if( longversion )
        {
            switch( type[0]*type[1] )
            {
            case 'S'*'A':
                // saddle
                sa = (struct saddle *)p;
                DUMP(( "   Epsilon = %g  original=%d", (float)(sa->epsilon), sa->dummy ))
                DUMP(( "   Transformation Matrix = [ [%g,%g], [%g,%g] ]", (float)(sa->a11),
                                        (float)(sa->a12), (float)(sa->a21), (float)(sa->a22) ))
                DUMP(( "   Vector Field:" ))
                DUMP(( "      x' = %s", DumpPoly2( sa->vector_field[0], "x", "y" ) ))
                DUMP(( "      y' = %s", DumpPoly2( sa->vector_field[1], "x", "y" ) ))
                DUMP(( "   Separatrices:" ))
                DumpSeparatrices( m, sa->separatrices, 6 );
                break;

            case 'D'*'E':
                // degenerate
                de = (struct degenerate *)p;
                DUMP(( "   Epsilon = %g  original=%d", (float)(de->epsilon), de->dummy ))
                break;

            case 'S'*'T':
                // strong focus
                sf = (struct strong_focus *)p;
                if( sf->stable == -1 ) s = "(stable)";
                else if( sf->stable == +1 ) s = "(unstable)";
                else s = "( ??? )";

                ss = s.toAscii();
                DUMP(( "    Stability = %d %s", sf->stable, (const char *)ss ))
                break;
            case 'W'*'E':
                // weak focus
                wf = (struct weak_focus *)p;
                DUMP(( "    Type = %d", wf->type ))
                break;
            case 'N'*'O':
                // node
                no = (struct node *)p;
                if( no->stable == -1 ) s = "(stable)";
                else if( no->stable == +1 ) s = "(unstable)";
                else s = "( ??? )";
                ss = s.toAscii();
                DUMP(( "    Stability = %d %s", no->stable, (const char *)ss ))
                break;
            case 'S'*'E':
                // semi-elementary
                se = (struct semi_elementary *)p;
                DUMP(( "   Type    = %d", se->type ))
                DUMP(( "   Epsilon = %g  original=%d", (float)(se->epsilon), se->dummy ))
                DUMP(( "   Transformation Matrix = [ [%g,%g], [%g,%g] ]", (float)(se->a11),
                                        (float)(se->a12), (float)(se->a21), (float)(se->a22) ))
                DUMP(( "   Vector Field = " ))
                DUMP(( "      x' = %s", DumpPoly2( se->vector_field[0], "x", "y" ) ))
                DUMP(( "      y' = %s", DumpPoly2( se->vector_field[1], "x", "y" ) ))
                DUMP(( "   Separatrices:" ))
                DumpSeparatrices( m, se->separatrices, 6 );
                break;
            }
            DUMP((" "))
        }
        p = p->next;
    }
}

void QVFStudy::Dump( QString basename )
{
    QTextEdit * m;
    QString s;
    QByteArray ss;

    m = new QTextEdit();
    
    DUMP(( "General Info" ))
    DUMP(( "------------" ))
    DUMP(( "  Type of study = %d", typeofstudy ))
    DUMP(( "  (p,q) = (%d,%d)", p, q ))
    DUMP(( "  plweights = %d", plweights ))
    DUMP(( "  doubles p,q,p+q,p-1,q-1,q-p = (%g,%g,%g,%g,%g,%g)", (float)double_p,
            (float)double_q, (float)double_p_plus_q, (float)double_p_minus_1,
            (float)double_q_minus_1, (float)double_q_minus_p ))
    DUMP(( "  Range x: [%g,%g] Range y: [%g,%g]", (float)xmin, (float)xmax, (float)ymin, (float)ymax ))
    DUMP(( "  Line at infinity singular? %d", singinf ))
    DUMP(( "  Direction of vector field: %d", VFResults.dir_vec_field ))
    DUMP(( " " ))
    DUMP(( "Vector Fields" ))
    DUMP(( "-------------" ))
    DUMP(( "  Finite chart:" ))
    DUMP(( "     x' = %s", DumpPoly2( f_vec_field[0], "x", "y" ) ))
    DUMP(( "     y' = %s", DumpPoly2( f_vec_field[1], "x", "y" ) ))
    DUMP(( "  U1 chart:" ))
    DUMP(( "     x' = %s", DumpPoly2( vec_field_U1[0], "x", "y" ) ))
    DUMP(( "     y' = %s", DumpPoly2( vec_field_U1[1], "x", "y" ) ))
    DUMP(( "  U2 chart:" ))
    DUMP(( "     x' = %s", DumpPoly2( vec_field_U2[0], "x", "y" ) ))
    DUMP(( "     y' = %s", DumpPoly2( vec_field_U2[1], "x", "y" ) ))
    DUMP(( "  V1 chart:" ))
    DUMP(( "     x' = %s", DumpPoly2( vec_field_V1[0], "x", "y" ) ))
    DUMP(( "     y' = %s", DumpPoly2( vec_field_V1[1], "x", "y" ) ))
    DUMP(( "  V2 chart:" ))
    DUMP(( "     x' = %s", DumpPoly2( vec_field_V2[0], "x", "y" ) ))
    DUMP(( "     y' = %s", DumpPoly2( vec_field_V2[1], "x", "y" ) ))
    DUMP(( "  Cylinder chart: (Co=cos(theta),Si=sin(theta))" ))
    DUMP(( "     r'     = %s", DumpPoly3( vec_field_C[0], "r", "Co", "Si" ) ))
    DUMP(( "     theta' = %s", DumpPoly3( vec_field_C[1], "r", "Co", "Si" ) ))
    DUMP(( " " ))
    DUMP(( "Greatest Common Factor" ))
    DUMP(( "----------------------" ))
    DUMP(( "  Finite chart:   %s", DumpPoly2( gcf, "x", "y" ) ))
    DUMP(( "  U1 chart:       %s", DumpPoly2( gcf_U1, "x", "y" ) ))
    DUMP(( "  U2 chart:       %s", DumpPoly2( gcf_U2, "x", "y" ) ))
    DUMP(( "  V1 chart:       %s", DumpPoly2( gcf_V1, "x", "y" ) ))
    DUMP(( "  V2 chart:       %s", DumpPoly2( gcf_V2, "x", "y" ) ))
    DUMP(( "  Cylinder chart: %s", DumpPoly3( gcf_C, "r", "Co", "Si" ) ))
    DUMP(( " " ))
    DUMP(( "Singular points - summary" ))
    DUMP(( "-------------------------" ))
    DUMP(( " " ))
    DumpSingularities( m, (struct genericsingularity *)first_saddle_point,  "SADDLE         ", false );
    DumpSingularities( m, (struct genericsingularity *)first_se_point,      "SEMI-ELEMENTARY", false );
    DumpSingularities( m, (struct genericsingularity *)first_node_point,    "NODE           ", false );
    DumpSingularities( m, (struct genericsingularity *)first_wf_point,      "WEAK FOCUS     ", false );
    DumpSingularities( m, (struct genericsingularity *)first_sf_point,      "STRONG FOCUS   ", false );
    DumpSingularities( m, (struct genericsingularity *)first_de_point,      "DEGENERATE     ", false );
    DUMP(( " " ))
    DUMP(( "Singular points - full description" ))
    DUMP(( "----------------------------------" ))
    DUMP(( " " ))
    DumpSingularities( m, (struct genericsingularity *)first_saddle_point,  "SADDLE         ", true );
    DumpSingularities( m, (struct genericsingularity *)first_se_point,      "SEMI-ELEMENTARY", true );
    DumpSingularities( m, (struct genericsingularity *)first_node_point,    "NODE           ", true );
    DumpSingularities( m, (struct genericsingularity *)first_wf_point,      "WEAK FOCUS     ", true );
    DumpSingularities( m, (struct genericsingularity *)first_sf_point,      "STRONG FOCUS   ", true );
    DumpSingularities( m, (struct genericsingularity *)first_de_point,      "DEGENERATE     ", true );
    DUMP(( " " ))
    DUMP(( "Default integration parameters" ))
    DUMP(( "------------------------------" ))
    DUMP(( " " ))
    DUMP(( "  LC Value           %d", config_lc_value ))
    DUMP(( "  Max Step Size      %g", (float)config_hma ))
    DUMP(( "  Min Step Size      %g", (float)config_hmi ))
    DUMP(( "  Step Size          %g", (float)config_step ))
    DUMP(( "  Current Step Size  %g", (float)config_currentstep ))
    DUMP(( "  Tolerance          %g", (float)config_tolerance ))
    DUMP(( "  Projection         %g", (float)config_projection ))
    DUMP(( "  Integration Points %d", config_intpoints ))
    DUMP(( "  LC NumPoints       %d", config_lc_numpoints ))
    DUMP(( "  Dashes             %d", config_dashes ))
    
    if( config_kindvf == INTCONFIG_ORIGINAL ) s = "(Original)";
        else if( config_kindvf == INTCONFIG_REDUCED ) s = "(Reduced)";
            else s = "???";

    ss = s.toAscii();
    DUMP(( "  Kind of VF         %d %s", config_kindvf, (const char *)ss ))

    ss= basename.toAscii() + "_fin.tab";
    QFile ffin( ss );
    if( ffin.open( QIODevice::ReadOnly ) )
    {
        DUMP(( " " ))
        DUMP(( "Contents of filename_fin.tab" ))
        DUMP(( "----------------------------" ))
        DUMP(( " " ))
        QTextStream tfin( &ffin );
        while( !tfin.atEnd() )
        {
            s = tfin.readLine();
            m->append( s );
        }
        ffin.close();
    }
    else
    {
        DUMP(( " " ))
        DUMP(( "Contents of filename_fin.tab" ))
        DUMP(( "----------------------------" ))
        DUMP(( " " ))
        DUMP(( "FILE DOES NOT EXIST." ))
    }
    QFile finf( basename + "_inf.tab" );
    if( finf.open( QIODevice::ReadOnly ) )
    {
        DUMP(( " " ))
        DUMP(( "Contents of filename_inf.tab" ))
        DUMP(( "----------------------------" ))
        DUMP(( " " ))

        QTextStream tinf( &finf );
        while( !tinf.atEnd() )
        {
            s = tinf.readLine();
            m->append( s );
        }
        finf.close();
    }
    else
    {
        DUMP(( " " ))
        DUMP(( "Contents of filename_inf.tab" ))
        DUMP(( "----------------------------" ))
        DUMP(( " " ))
        DUMP(( "FILE DOES NOT EXIST." ))
    }
    DUMP(( " " ))

    m->setFont( *(p4app->CourierFont) );
    m->setReadOnly(true);
    m->resize(640,480);
    SetP4WindowTitle( m, "DUMP SCREEN" );
    m->show();
}

