// -----------------------------------------------------------------------
//
//                      NUMERIC ROUTINES
//
// - bisection
// - regula falsi
// - newton method to find a root
// - Runge-Kutta integration 7/8
//
// -----------------------------------------------------------------------

#include <math.h>
#include <float.h>

#define p4_isnan        isnan       // used to be this function
#define p4_finite       finite

static double PRECISION1 = 1e-16;
static double PRECISION2 = 1e-8;

// -----------------------------------------------------------------------
//                              BISECTION
// -----------------------------------------------------------------------
//
// dx is a pointer to an array of two elements

void bisection( double (* f)(double), double * x, double e )
{
    double fx0, fmid, xmid;

    if( (*f)( x[0] ) > 0 )
    {
        xmid = x[0];
        x[0] = x[1];
        x[1] = xmid;
    }

    for( ;; )
    {
        xmid = (x[0] + x[1])/2;

        fx0 = (*f)( x[0] );
        fmid = (*f)( xmid );

        if( fx0 * fmid < 0)
            x[1] = xmid;
        else
            x[0] = xmid;

        if( fabs( x[1] - x[0] ) < e )
            break;
    }
}

// -----------------------------------------------------------------------
//                          REGULA FALSI
// -----------------------------------------------------------------------
//
// x is an array of two elements

double regula_falsi( double (*f)(double), double * x, double e )
{
    double x2;
    double y;

    x2 = 2*x[1] - x[0];

    for(;;)
    {
        y = x[1] - (*f)(x[1]) * ((x[1] - x[0])/((*f)(x[1]) - (*f)(x[0])));

        if( fabs( y - x2 ) < e && f(y) < PRECISION2 )
            break;

        if( (*f)(x[1]) * (*f)(y) <= 0 )
            x[0] = y;
        else
            x[1] = y;

        x2 = y;
    }

    return y;
}

// -----------------------------------------------------------------------
//                              NEWTON
// -----------------------------------------------------------------------

double newton( double (*f)(double), double (*df)(double), double x, double e )
{
    double dx;

    if( fabs( (*f)(x) ) < PRECISION1 && fabs( (*f)(x) / (*df)(x) ) < e )
        return x;

    for(;;)
    {
        dx = (*f)(x) / (*df)(x);
        x -= dx;
        if( fabs( (*f)(x) ) < PRECISION1 || fabs(dx) < e )
            break;
    }

    return x;
}

// -----------------------------------------------------------------------
//                              FIND_ROOT
// -----------------------------------------------------------------------
//
// value is a pointer to an array of two elements

double find_root( double (*f)(double), double (*df)(double), double * value )
{
    double y;

    bisection( f, value, 0.01 );
    y = regula_falsi( f, value, 0.01);
    y = newton(f, df, y, 1e-8 );

    return y;
}

// -----------------------------------------------------------------------
//                              RK78
// -----------------------------------------------------------------------

static const double __c[11] =
{
    .04880952380952381,
    0.,
    0.,
    0.,
    0.,
    .32380952380952382,
    .25714285714285712,
    .25714285714285712,
    .03214285714285714,
    .03214285714285714,
    .03214285714285714
};

static const double __beta[79] =
{
    0.,
    .07407407407407407,
    .027777777777777776,
    .083333333333333329,
    .041666666666666664,
    0.,
    .125,
    .41666666666666669,
    0.,
    -1.5625,
    1.5625,
    .05,
    0.,
    0.,
    .25,
    .2,
    -.23148148148148148,
    0.,
    0.,
    1.1574074074074074,
    -2.4074074074074074,
    2.3148148148148148,
    .10333333333333333,
    0.,
    0.,
    0.,
    .27111111111111114,
    -.22222222222222221,
    .014444444444444444,
    2.,
    0.,
    0.,
    -8.8333333333333339,
    15.644444444444444,
    -11.888888888888889,
    .74444444444444446,
    3.,
    -.84259259259259256,
    0.,
    0.,
    .21296296296296297,
    -7.2296296296296294,
    5.7592592592592595,
    -.31666666666666665,
    2.8333333333333335,
    -.083333333333333329,
    .58121951219512191,
    0.,
    0.,
    -2.0792682926829267,
    4.3863414634146345,
    -3.6707317073170733,
    .52024390243902441,
    .54878048780487809,
    .27439024390243905,
    .43902439024390244,
    .014634146341463415,
    0.,
    0.,
    0.,
    0.,
    -.14634146341463414,
    -.014634146341463415,
    -.073170731707317069,
    .073170731707317069,
    .14634146341463414,
    0.,
    -.43341463414634146,
    0.,
    0.,
    -2.0792682926829267,
    4.3863414634146345,
    -3.524390243902439,
    .53487804878048784,
    .62195121951219512,
    .20121951219512196,
    .29268292682926828,
    0.,
    1.
};

void rk78( void (*deriv)( const double *, double * ), double y[2], double * hh,
            double hmi, double hma, double e1)
{
    double d,dd,e3,h,r[13][2],b[2],f[2];
    int k;
    int direction;

    h = *hh;
    if( h < 0 )
        direction = -1;
    else
        direction = 1;

    for(;;)
    {
        b[0]=y[0];
        b[1]=y[1];
        deriv(b,r[0]);

        b[0]=y[0]+__beta[1]*r[0][0]*h;
        b[1]=y[1]+__beta[1]*r[0][1]*h;
        deriv(b,r[1]);

        b[0]=y[0]+(__beta[2]*r[0][0]+__beta[3]*r[1][0])*h;
        b[1]=y[1]+(__beta[2]*r[0][1]+__beta[3]*r[1][1])*h;
        deriv(b,r[2]);

        b[0]=y[0]+(__beta[4]*r[0][0]+__beta[6]*r[2][0])*h;
        b[1]=y[1]+(__beta[4]*r[0][1]+__beta[6]*r[2][1])*h;
        deriv(b,r[3]);

        b[0]=y[0]+(__beta[7]*r[0][0]+__beta[9]*(r[2][0]-r[3][0]))*h;
        b[1]=y[1]+(__beta[7]*r[0][1]+__beta[9]*(r[2][1]-r[3][1]))*h;
        deriv(b,r[4]);

        for(k=0;k<2;++k)
            b[k]=y[k]+(__beta[11]*r[0][k]+__beta[14]*r[3][k]+__beta[15]*r[4][k])*h;
        deriv(b,r[5]);

        for(k=0;k<2;++k)
            b[k]=y[k]+(__beta[16]*r[0][k]+__beta[19]*r[3][k]+__beta[20]*r[4][k]+__beta[21]*r[5][k])*h;
        deriv(b,r[6]);

        for(k=0;k<2;++k)
            b[k]=y[k]+(__beta[22]*r[0][k]+__beta[26]*r[4][k]+__beta[27]*r[5][k]+__beta[28]*r[6][k])*h;
        deriv(b,r[7]);

        for(k=0;k<2;++k)
            b[k]=y[k]+(__beta[29]*r[0][k]+__beta[32]*r[3][k]+__beta[33]*r[4][k]+__beta[34]*r[5][k]+__beta[35]*r[6][k]+__beta[36]*r[7][k])*h;
        deriv(b,r[8]);

        for(k=0;k<2;++k)
            b[k]=y[k]+(__beta[37]*r[0][k]+__beta[40]*r[3][k]+__beta[41]*r[4][k]+__beta[42]*r[5][k]+__beta[43]*r[6][k]+__beta[44]*r[7][k]+__beta[45]*r[8][k])*h;
        deriv(b,r[9]);

        for(k=0;k<2;++k)
            b[k]=y[k]+(__beta[46]*r[0][k]+__beta[49]*r[3][k]+__beta[50]*r[4][k]+__beta[51]*r[5][k]+__beta[52]*r[6][k]+__beta[53]*r[7][k]+__beta[54]*r[8][k]+__beta[55]*r[9][k])*h;
        deriv(b,r[10]);

        for(k=0;k<2;++k)
            b[k]=y[k]+(__beta[56]*r[0][k]+__beta[61]*(r[5][k]-r[9][k])+__beta[62]*r[6][k]+__beta[63]*(r[7][k]-r[8][k]))*h;
        deriv(b,r[11]);

        for(k=0;k<2;++k)
            b[k]=y[k]+(__beta[67]*r[0][k]+__beta[70]*r[3][k]+__beta[71]*r[4][k]+__beta[72]*r[5][k]+__beta[73]*r[6][k]+__beta[74]*r[7][k]+__beta[75]*r[8][k]+__beta[76]*r[9][k]+r[11][k])*h;
        deriv(b,r[12]);

        d=0;
        dd=0;

        for(k=0;k<2;++k)
        {
            b[k]=__c[5]*r[5][k]+__c[6]*(r[6][k]+r[7][k])+__c[8]*(r[8][k]+r[9][k]);
            f[k]=y[k]+h*(b[k]+__c[0]*(r[11][k]+r[12][k]));
            b[k]=y[k]+h*(b[k]+__c[0]*(r[0][k]+r[10][k]));
            d=d+fabs(f[k]-b[k]);
            dd=dd+fabs(f[k]);
        }
        d=d/2;
        e3=e1*(1.0+dd*1.E-2);
        if( (fabs(h)<=hmi)||(d<e3) && !p4_isnan(f[0]) && !p4_isnan(f[1]) && p4_finite(f[0]) && p4_finite(f[1]) )
            break;

        h = h*0.9*sqrt(sqrt(sqrt(e3/d)));

        if((fabs(h)<hmi) || p4_isnan(h) || !p4_finite(h) )  /* h=hmi*h/fabs(h); */
            h=hmi*direction;
    }
    if( d < e3 / 512 )
        d = e3 / 512;

    h = h * 0.9 * sqrt(sqrt(sqrt(e3/d)));
    for(k = 0; k < 2; ++k )
        y[k]=f[k];

    if( fabs(h) > hma )
        h = hma*direction;

    if((fabs(h)<hmi) || p4_isnan(h) || !p4_finite(h) )
        h = hmi*direction;
    *hh = h;
}
