#include <qfile.h>
#include <math.h>
#include "table.h"
#include "custom.h"
#include "file_vf.h"
#include "win_p5.h"
#include "math_p4.h"
#include "win_inputsphere.h"
/*

    In this file, routines are developped to check if a point belongs
    to a given region, and to find the associated vector field.


    In order to avoid calculations with large numbers, when a point
    is close to infinity, the calculations are not done in usual
    coordinates, but instead in one of the charts near infinity.

    Therefore, a lot of variants have been developed with care.

*/

// ---------------------------------------------------------------------
//                      GETVFINDEX_R2
// ---------------------------------------------------------------------
//
// Given a coordinate ucoord=(x0,y0), we find the index of the vector
// field that is assigned to the region where the point belongs to.
//
// The point is given in standard coordinates (in the finite region)
// and is assumed to be in a ball with small radius.

int QInputVF::GetVFIndex_R2( const double * ucoord )
{
    if( VFResults.curves == NULL ) return -1;
    for( int i = numvfregions-1; i >= 0; i-- )
    {
        if( IsInsideRegion_R2( vfregions[i].signs, ucoord ) )
            return vfregions[i].vfindex;
    }
    return -1;
}

// ---------------------------------------------------------------------
//                      GETVFINDEX_SPHERE
// ---------------------------------------------------------------------
//
// Given a coordinate pcoord=(X,Y,Z), we find the index of the vector
// field that is assigned to the region where the point belongs to.
//
// The point is given in Poincare-sphere coordinates (X,Y,Z), or
// Poincare-Lyapunov coordinates (X,Y,Z) depending on the setting of the
// system

int QInputVF::GetVFIndex_sphere( const double * pcoord )
{
    double ucoord[2];

    if( !VFResults.plweights )
    {
        double theta;

        if( pcoord[2]>ZCOORD )
        {
            psphere_to_R2( pcoord[0],pcoord[1],pcoord[2], ucoord );
            return GetVFIndex_R2( ucoord );
        }
        else
        {
            theta=atan2(fabs(pcoord[1]),fabs(pcoord[0]));
            if( (theta<PI_DIV4) && (theta>-PI_DIV4) )
            {
                if( pcoord[0] > 0 )
                {
                    psphere_to_U1( pcoord[0],pcoord[1],pcoord[2], ucoord );
                    return GetVFIndex_U1( ucoord );
                }
                else
                {
                    psphere_to_V1( pcoord[0],pcoord[1],pcoord[2], ucoord );
                    return GetVFIndex_V1( ucoord );
                }
            }
            else
            {
                if( pcoord[1] > 0 )
                {
                    psphere_to_U2( pcoord[0],pcoord[1],pcoord[2], ucoord );
                    return GetVFIndex_U2( ucoord );
                }
                else
                {
                    psphere_to_V2( pcoord[0],pcoord[1],pcoord[2], ucoord );
                    return GetVFIndex_V2( ucoord );
                }
            }
        }
    }
    else
    {
        ucoord[0] = pcoord[1];
        ucoord[1] = pcoord[2];

        if( pcoord[0] )
            return GetVFIndex_R2( ucoord );
        else
            return GetVFIndex_cyl( ucoord );
    }
}

// ---------------------------------------------------------------------
//                      GETVFINDEX_PSPHERE
// ---------------------------------------------------------------------
//
// Given a coordinate pcoord=(X,Y,Z), we find the index of the vector
// field that is assigned to the region where the point belongs to.
//
// The point is given in Poincare-sphere coordinates (X,Y,Z), where
// (X,Y) lies on the unit circle and Z is (the inverse of) a radial coordinate.
//
// Since we do not know if the point is close to infinity or not, and
// if it is, in which chart it lies, we need to first find that out.
//
// A copy of this routine is pasted in a part the routine GetVFIndex_sphere.

int QInputVF::GetVFIndex_psphere( const double * pcoord )
{
    double ucoord[2];
    double theta;

    if( pcoord[2]>ZCOORD )
    {
        psphere_to_R2( pcoord[0],pcoord[1],pcoord[2], ucoord );
        return GetVFIndex_R2( ucoord );
    }
    else
    {
        theta=atan2(fabs(pcoord[1]),fabs(pcoord[0]));
        if( (theta<PI_DIV4) && (theta>-PI_DIV4) )
        {
            if( pcoord[0] > 0 )
            {
                psphere_to_U1( pcoord[0],pcoord[1],pcoord[2], ucoord );
                return GetVFIndex_U1( ucoord );
            }
            else
            {
                psphere_to_V1( pcoord[0],pcoord[1],pcoord[2], ucoord );
                return GetVFIndex_V1( ucoord );
            }
        }
        else
        {
            if( pcoord[1] > 0 )
            {
                psphere_to_U2( pcoord[0],pcoord[1],pcoord[2], ucoord );
                return GetVFIndex_U2( ucoord );
            }
            else
            {
                psphere_to_V2( pcoord[0],pcoord[1],pcoord[2], ucoord );
                return GetVFIndex_V2( ucoord );
            }
        }
    }
}

// ---------------------------------------------------------------------
//                      GETVFINDEX_PLSPHERE
// ---------------------------------------------------------------------
//
// Given a coordinate pcoord=(X,Y,Z), we find the index of the vector
// field that is assigned to the region where the point belongs to.
//
// The point is given in Poincare-sphere coordinates (X,Y,Z).  Such coordinates
// can be of two forms:
//
//      (X,Y,Z) = (0,u,v)  --> (x,y) = (u,v)
//      (X,Y,Z) = (1,u,v)  --> (x,y) = (cos(v)/u^p, sin(v)/u^q)
//
// The first form is used for points inside the unit circle; the second
// for points near infinity.  In the above form, the relation with the
// normal (x,y)-coordinates in the plane are shown.
//
// A copy of this routine is pasted in a part the routine GetVFIndex_sphere.

int QInputVF::GetVFIndex_plsphere( const double * pcoord )
{
    double ucoord[2];

    ucoord[0] = pcoord[1];
    ucoord[1] = pcoord[2];

    if( pcoord[0] )
        return GetVFIndex_R2( ucoord );
    else
        return GetVFIndex_cyl( ucoord );
}

// ---------------------------------------------------------------------
//                      GETVFINDEX_CYL
// ---------------------------------------------------------------------
//
// Given a coordinate y=(r,theta), we find the index of the vector
// field that is assigned to the region where the point belongs to.
//
// This is only used in the Poincare-Lyapunov setting, in a study
// near infinity.
//
// The point is given in cylindrical coordinates (r,theta).  The r-coordinate
// is a (inverse) radial coordinate, the theta is an angular coordinate.
//

int QInputVF::GetVFIndex_cyl( const double * y )
{
    if( VFResults.curves == NULL ) return -1;
    for( int i = numvfregions-1; i >= 0; i-- )
    {
        if( IsInsideRegion_cyl( vfregions[i].signs, y ) )
            return vfregions[i].vfindex;
    }
    return -1;
}

// ---------------------------------------------------------------------
//                      GETVFINDEX_U1
// ---------------------------------------------------------------------
//
// Given a coordinate y=(z1,z2), we find the index of the vector
// field that is assigned to the region where the point belongs to.
//
// The relation with original coordinates is:
//
//          (x,y) = (1/z2,z1/z2),           z2 > 0.
//
// If z2<0, we are inside the chart VV1.

int QInputVF::GetVFIndex_U1( const double * y )
{
    if( VFResults.curves == NULL ) return -1;
    for( int i = numvfregions-1; i >= 0; i-- )
    {
        if( IsInsideRegion_U1( vfregions[i].signs, y ) )
            return vfregions[i].vfindex;
    }
    return -1;
}

// ---------------------------------------------------------------------
//                      GETVFINDEX_V1
// ---------------------------------------------------------------------
//
// Given a coordinate y=(z1,z2), we find the index of the vector
// field that is assigned to the region where the point belongs to.
//
// The relation with original coordinates is:
//
//          (x,y) = (-1/z2,z1/z2),           z2 > 0.
//
// If z2<0, we are inside the chart UU1.

int QInputVF::GetVFIndex_V1( const double * y )  // only used when plweights=1
{
    if( VFResults.curves == NULL ) return -1;
    for( int i = numvfregions-1; i >= 0; i-- )
    {
        if( IsInsideRegion_V1( vfregions[i].signs, y ) )
            return vfregions[i].vfindex;
    }
    return -1;
}

// ---------------------------------------------------------------------
//                      GETVFINDEX_U2
// ---------------------------------------------------------------------
//
// Given a coordinate y=(z1,z2), we find the index of the vector
// field that is assigned to the region where the point belongs to.
//
// The relation with original coordinates is:
//
//          (x,y) = (z1/z2,1/z2),           z2 > 0.
//
// If z2<0, we are inside the chart VV2.

int QInputVF::GetVFIndex_U2( const double * y )
{
    if( VFResults.curves == NULL ) return -1;
    for( int i = numvfregions-1; i >= 0; i-- )
    {
        if( IsInsideRegion_U2( vfregions[i].signs, y ) )
            return vfregions[i].vfindex;
    }
    return -1;
}

// ---------------------------------------------------------------------
//                      GETVFINDEX_V2
// ---------------------------------------------------------------------
//
// Given a coordinate y=(z1,z2), we find the index of the vector
// field that is assigned to the region where the point belongs to.
//
// The relation with original coordinates is:
//
//          (x,y) = (z1/z2,-1/z2),           z2 > 0.
//
// If z2<0, we are inside the chart UU2.

int QInputVF::GetVFIndex_V2( const double * y )
{
    if( VFResults.curves == NULL ) return -1;
    for( int i = numvfregions-1; i >= 0; i-- )
    {
        if( IsInsideRegion_V2( vfregions[i].signs, y ) )
            return vfregions[i].vfindex;
    }
    return -1;
}

// ---------------------------------------------------------------------
//                      GETVFINDEX_UU1
//                      GETVFINDEX_UU2
//                      GETVFINDEX_VV1
//                      GETVFINDEX_VV2
// ---------------------------------------------------------------------
//
// When integration starts, for example in a U1 chart in a point (z1,z2)
// very close to infinity (close to z2=0, but z2>0), it may sometimes happen
// that the line of infinity is crossed (to a point z2<0).  The resulting
// point lies in the U1 chart, but in a wrong part of it.  We say that
// it lies in the VV1-part: this point will be mapped to the symmetric point
// in the V1-chart.
//
// So:
//
//   * a point in the UU1 chart lies in the wrong part of the V1-chart, and is in fact a U1-point
//   * a point in the UU2 chart lies in the wrong part of the V2-chart, and is in fact a U2-point
//   * a point in the VV1 chart lies in the wrong part of the U1-chart, and is in fact a V1-point
//   * a point in the VV2 chart lies in the wrong part of the U2-chart, and is in fact a V2-point
//
// The relation between the symmetric charts is simple:
//              (z1,z2) ---> (-z1,-z2).
//
// Note: these charts are only used when p=q=1.

int QInputVF::GetVFIndex_UU1( const double * yy )
{
    double y[2];
    y[0] = -yy[0];
    y[1] = -yy[1];
    return GetVFIndex_U1( y );
}

int QInputVF::GetVFIndex_UU2( const double * yy )
{
    double y[2];
    y[0] = -yy[0];
    y[1] = -yy[1];
    return GetVFIndex_U2( y );
}

int QInputVF::GetVFIndex_VV1( const double * yy )
{
    double y[2];
    y[0] = -yy[0];
    y[1] = -yy[1];
    return GetVFIndex_V1( y );
}

int QInputVF::GetVFIndex_VV2( const double * yy )
{
    double y[2];
    y[0] = -yy[0];
    y[1] = -yy[1];
    return GetVFIndex_V2( y );
}

// ---------------------------------------------------------------------
//                      ISINSIDEREGION_R2
//                      ISINSIDEREGION_R2_EPSILON
// ---------------------------------------------------------------------
//
// Checks if a points lies in a region determined by a set of signs.
// The point lies in the finite region.
//
// The _epsilon variant increases the region by an epsilon amount before checking
//

bool IsInsideRegion_R2( const int * signs, const double * ucoord )
{
    int k;

    if( VFResults.curves == NULL )
        return false;

    for( k = ThisVF->numcurves-1; k >= 0; k-- )
    {
        if( eval_term2( VFResults.curves[k].sep, ucoord ) < 0 )
        {
            if( signs[k] > 0 )
                return false;
        }
        else
        {
            if( signs[k] < 0 )
                return false;
        }
    }
    return true;
}
bool IsInsideRegion_R2_epsilon( const int * signs, const double * ucoord, double epsilon )
{
    int k;
    double v;

    if( VFResults.curves == NULL )
        return false;

    for( k = ThisVF->numcurves-1; k >= 0; k-- )
    {
        v = eval_term2( VFResults.curves[k].sep, ucoord );
        if( v < -epsilon )
        {
            if( signs[k] > 0 )
                return false;
        }
        if( v > epsilon )
        {
            if( signs[k] < 0 )
                return false;
        }
    }
    return true;
}

// ---------------------------------------------------------------------
//                      ISINSIDEREGION_U1
//                      ISINSIDEREGION_U1_EPSILON
// ---------------------------------------------------------------------
//
// Checks if a points lies in a region determined by a set of signs.
// The point lies in the U1 chart.
// This is only used when p=q=1
//
// The _epsilon variant increases the region by an epsilon amount before checking
//

bool IsInsideRegion_U1( const int * signs, const double * z1z2 )
{
    int k;

    if( VFResults.curves == NULL )
        return false;

    for( k = ThisVF->numcurves-1; k >= 0; k-- )
    {
        if( eval_term2( VFResults.curves[k].sep_U1, z1z2 ) < 0 )
        {
            if( signs[k] > 0 )
                return false;
        }
        else
        {
            if( signs[k] < 0 )
                return false;
        }
    }
    return true;
}
bool IsInsideRegion_U1_epsilon( const int * signs, const double * z1z2, double epsilon )
{
    int k;
    double v;

    if( VFResults.curves == NULL )
        return false;

    for( k = ThisVF->numcurves-1; k >= 0; k-- )
    {
        v = eval_term2( VFResults.curves[k].sep_U1, z1z2 );
        if( v < -epsilon )
        {
            if( signs[k] > 0 )
                return false;
        }
        if( v > epsilon )
        {
            if( signs[k] < 0 )
                return false;
        }
    }
    return true;
}

// ---------------------------------------------------------------------
//                      ISINSIDEREGION_V1
//                      ISINSIDEREGION_V1_EPSILON
// ---------------------------------------------------------------------
//
// Checks if a points lies in a region determined by a set of signs.
// The point lies in the V1 chart.
// This is only used when p=q=1
//
// The _epsilon variant increases the region by an epsilon amount before checking
//

bool IsInsideRegion_V1( const int * signs, const double * z1z2 )
{
    int k;

    if( VFResults.curves == NULL )
        return false;

    for( k = ThisVF->numcurves-1; k >= 0; k-- )
    {
        if( eval_term2( VFResults.curves[k].sep_V1, z1z2 ) < 0 )
        {
            if( signs[k] > 0 )
                return false;
        }
        else
        {
            if( signs[k] < 0 )
                return false;
        }
    }
    return true;
}
bool IsInsideRegion_V1_epsilon( const int * signs, const double * z1z2, double epsilon )
{
    int k;
    double v;

    if( VFResults.curves == NULL )
        return false;

    for( k = ThisVF->numcurves-1; k >= 0; k-- )
    {
        v = eval_term2( VFResults.curves[k].sep_V1, z1z2 );
        if( v < -epsilon )
        {
            if( signs[k] > 0 )
                return false;
        }
        if( v > epsilon )
        {
            if( signs[k] < 0 )
                return false;
        }
    }
    return true;
}

// ---------------------------------------------------------------------
//                      ISINSIDEREGION_U2
//                      ISINSIDEREGION_U2_EPSILON
// ---------------------------------------------------------------------
//
// Checks if a points lies in a region determined by a set of signs.
// The point lies in the U2 chart.
// This is only used when p=q=1
//
// The _epsilon variant increases the region by an epsilon amount before checking
//

bool IsInsideRegion_U2( const int * signs, const double * z1z2 )
{
    int k;

    if( VFResults.curves == NULL )
        return false;

    for( k = ThisVF->numcurves-1; k >= 0; k-- )
    {
        if( eval_term2( VFResults.curves[k].sep_U2, z1z2 ) < 0 )
        {
            if( signs[k] > 0 )
                return false;
        }
        else
        {
            if( signs[k] < 0 )
                return false;
        }
    }
    return true;
}
bool IsInsideRegion_U2_epsilon( const int * signs, const double * z1z2, double epsilon )
{
    int k;
    double v;

    if( VFResults.curves == NULL )
        return false;

    for( k = ThisVF->numcurves-1; k >= 0; k-- )
    {
        v = eval_term2( VFResults.curves[k].sep_U2, z1z2 );
        if( v < -epsilon )
        {
            if( signs[k] > 0 )
                return false;
        }
        if( v > epsilon )
        {
            if( signs[k] < 0 )
                return false;
        }
    }
    return true;
}

// ---------------------------------------------------------------------
//                      ISINSIDEREGION_V2
//                      ISINSIDEREGION_V2_EPSILON
// ---------------------------------------------------------------------
//
// Checks if a points lies in a region determined by a set of signs.
// The point lies in the V2 chart.
// This is only used when p=q=1
//
// The _epsilon variant increases the region by an epsilon amount before checking
//

bool IsInsideRegion_V2( const int * signs, const double * z1z2 )
{
    int k;

    if( VFResults.curves == NULL )
        return false;

    for( k = ThisVF->numcurves-1; k >= 0; k-- )
    {
        if( eval_term2( VFResults.curves[k].sep_V2, z1z2 ) < 0 )
        {
            if( signs[k] > 0 )
                return false;
        }
        else
        {
            if( signs[k] < 0 )
                return false;
        }
    }
    return true;
}
bool IsInsideRegion_V2_epsilon( const int * signs, const double * z1z2, double epsilon )
{
    int k;
    double v;

    if( VFResults.curves == NULL )
        return false;

    for( k = ThisVF->numcurves-1; k >= 0; k-- )
    {
        v = eval_term2( VFResults.curves[k].sep_V2, z1z2 );
        if( v < -epsilon )
        {
            if( signs[k] > 0 )
                return false;
        }
        if( v > epsilon )
        {
            if( signs[k] < 0 )
                return false;
        }
    }
    return true;
}

// ---------------------------------------------------------------------
//                      ISINSIDEREGION_CYL
//                      ISINSIDEREGION_CYL_EPSILON
// ---------------------------------------------------------------------
//
// Checks if a points lies in a region determined by a set of signs.
// The point lies in the part near infinity, in the case of Poincare-
// Lyapunov compactification ((p,q) is not (1,1)).
//
// The _epsilon variant increases the region by an epsilon amount before checking
//

bool IsInsideRegion_cyl( const int * signs, const double * rtheta )
{
    int k;

    if( VFResults.curves == NULL )
        return false;

    for( k = ThisVF->numcurves-1; k >= 0; k-- )
    {
        if( eval_term3( VFResults.curves[k].sep_C, rtheta ) < 0 )
        {
            if( signs[k] > 0 )
                return false;
        }
        else
        {
            if( signs[k] < 0 )
                return false;
        }
    }
    return true;
}
bool IsInsideRegion_cyl_epsilon( const int * signs, const double * rtheta, double epsilon )
{
    int k;
    double v;

    if( VFResults.curves == NULL )
        return false;

    for( k = ThisVF->numcurves-1; k >= 0; k-- )
    {
        v = eval_term3( VFResults.curves[k].sep_C, rtheta );
        if( v < -epsilon )
        {
            if( signs[k] > 0 )
                return false;
        }
        if( v > epsilon )
        {
            if( signs[k] < 0 )
                return false;
        }
    }
    return true;
}
// ---------------------------------------------------------------------
//                      ISINSIDEREGION_SPHERE
//                      ISINSIDEREGION_SPHERE_EPSILON
// ---------------------------------------------------------------------
//
// Checks if a points lies in a region determined by a set of signs.
// We don't know where the point lies, so we have to determine this first.
//
// The _epsilon variant increases the region by an epsilon amount before checking
//
bool IsInsideRegion_sphere( const int * signs, const double * pcoord )
{
    double ucoord[2];
    double theta;

    if( !VFResults.plweights )
    {
        if( pcoord[2]>ZCOORD )
        {
            psphere_to_R2( pcoord[0],pcoord[1],pcoord[2], ucoord );
            return IsInsideRegion_R2( signs, ucoord );
        }
        else
        {
            theta=atan2(fabs(pcoord[1]),fabs(pcoord[0]));
            if( (theta<PI_DIV4) && (theta>-PI_DIV4) )
            {
                if( pcoord[0] > 0 )
                {
                    psphere_to_U1( pcoord[0],pcoord[1],pcoord[2], ucoord );
                    return IsInsideRegion_U1( signs, ucoord );
                }
                else
                {
                    psphere_to_V1( pcoord[0],pcoord[1],pcoord[2], ucoord );
                    return IsInsideRegion_V1( signs, ucoord );
                }
            }
            else
            {
                if( pcoord[1] > 0 )
                {
                    psphere_to_U2( pcoord[0],pcoord[1],pcoord[2], ucoord );
                    return IsInsideRegion_U2( signs, ucoord );
                }
                else
                {
                    psphere_to_V2( pcoord[0],pcoord[1],pcoord[2], ucoord );
                    return IsInsideRegion_V2( signs, ucoord );
                }
            }
        }
    }
    else
    {
        ucoord[0] = pcoord[1];
        ucoord[1] = pcoord[2];

        if( pcoord[0] )
            return IsInsideRegion_R2( signs, ucoord );
        else
            return IsInsideRegion_cyl( signs, ucoord );
    }
}

bool IsInsideRegion_sphere_epsilon( const int * signs, const double * pcoord, double epsilon )
{
    double ucoord[2];
    double theta;

    if( !VFResults.plweights )
    {
        if( pcoord[2]>ZCOORD )
        {
            psphere_to_R2( pcoord[0],pcoord[1],pcoord[2], ucoord );
            return IsInsideRegion_R2_epsilon( signs, ucoord, epsilon );
        }
        else
        {
            theta=atan2(fabs(pcoord[1]),fabs(pcoord[0]));
            if( (theta<PI_DIV4) && (theta>-PI_DIV4) )
            {
                if( pcoord[0] > 0 )
                {
                    psphere_to_U1( pcoord[0],pcoord[1],pcoord[2], ucoord );
                    return IsInsideRegion_U1_epsilon( signs, ucoord, epsilon );
                }
                else
                {
                    psphere_to_V1( pcoord[0],pcoord[1],pcoord[2], ucoord );
                    return IsInsideRegion_V1_epsilon( signs, ucoord, epsilon );
                }
            }
            else
            {
                if( pcoord[1] > 0 )
                {
                    psphere_to_U2( pcoord[0],pcoord[1],pcoord[2], ucoord );
                    return IsInsideRegion_U2_epsilon( signs, ucoord, epsilon );
                }
                else
                {
                    psphere_to_V2( pcoord[0],pcoord[1],pcoord[2], ucoord );
                    return IsInsideRegion_V2_epsilon( signs, ucoord, epsilon );
                }
            }
        }
    }
    else
    {
        ucoord[0] = pcoord[1];
        ucoord[1] = pcoord[2];

        if( pcoord[0] )
            return IsInsideRegion_R2_epsilon( signs, ucoord, epsilon );
        else
            return IsInsideRegion_cyl_epsilon( signs, ucoord, epsilon );
    }
}

// ---------------------------------------------------------------------
//                      EVAL_CURVE
// ---------------------------------------------------------------------
//
// A polynom is evaluated in a point pcoord=(X,Y,Z).  Only the sign of
// the resulting polynomial is important, so therefore we can replace
// the polynomial by a rescaled one if the point lies close to infinity.
//
// So we first determine where the point lies.

double eval_curve( const CURVERESULT * c, const double * pcoord )
{
    double ucoord[2];
    double theta;

    if( !VFResults.plweights )
    {
        if( pcoord[2]>ZCOORD )
        {
            psphere_to_R2( pcoord[0],pcoord[1],pcoord[2], ucoord );
            return eval_term2( c->sep, ucoord );
        }
        else
        {
            theta=atan2(fabs(pcoord[1]),fabs(pcoord[0]));
            if( (theta<PI_DIV4) && (theta>-PI_DIV4) )
            {
                if( pcoord[0] > 0 )
                {
                    psphere_to_U1( pcoord[0],pcoord[1],pcoord[2], ucoord );
                    return eval_term2( c->sep_U1, ucoord );
                }
                else
                {
                    psphere_to_V1( pcoord[0],pcoord[1],pcoord[2], ucoord );
                    return eval_term2( c->sep_V1, ucoord );
                }
            }
            else
            {
                if( pcoord[1] > 0 )
                {
                    psphere_to_U2( pcoord[0],pcoord[1],pcoord[2], ucoord );
                    return eval_term2( c->sep_U2, ucoord );
                }
                else
                {
                    psphere_to_V2( pcoord[0],pcoord[1],pcoord[2], ucoord );
                    return eval_term2( c->sep_V2, ucoord );
                }
            }
        }
    }
    else
    {
        ucoord[0] = pcoord[1];
        ucoord[1] = pcoord[2];

        if( pcoord[0] )
            return eval_term2( c->sep, ucoord );
        else
            return eval_term3( c->sep_C, ucoord );
    }
}

// ---------------------------------------------------------------------
//                      DESCRIBEREGION
// ---------------------------------------------------------------------
//
// included for debugging region

QString DescribeRegion( double * pcoord )
{
    QString s;
    int k;
    double v;

    s = "";
    for( k = 0; k < ThisVF->numcurves; k++ )
    {
        v = eval_curve( VFResults.curves+k, pcoord );
        if( fabs(v) <= 1.e-8 )
            s += "0";
        else if( v < 0 )
            s += "-";
        else
            s += "+";
    }
    return s;
}

// ---------------------------------------------------------------------
//                      ISSINGULARITYINSAMEREGION
// ---------------------------------------------------------------------

bool IsInTheSameRegion( double * testpt, double * refpos )
{
    int i;
    if( refpos == NULL || VFResults.curves == NULL )
        return true;
        
    for( i = ThisVF->numcurves-1; i >= 0; i-- )
    {
        if( eval_curve( VFResults.curves+i, refpos ) < 0 )
        {
            if( eval_curve( VFResults.curves+i, testpt ) > 1e-8 )
                return false;
        }
        else
        {
            if( eval_curve( VFResults.curves+i, testpt ) < -1e-8 )
                return false;
        }
    }
    return true;
}

// ---------------------------------------------------------------------
//                      ISAREALSINGULARITY
// ---------------------------------------------------------------------

bool IsARealSingularity( double * pcoord, int vfindex )
{
    if( VFResults.curves == NULL ) return false;
    
    for( int i = ThisVF->numvfregions-1; i >= 0; i-- )
    {
        if( ThisVF->vfregions[i].vfindex == vfindex )
            if( IsInsideRegion_sphere_epsilon( ThisVF->vfregions[i].signs, pcoord, 1e-8 ) )
                return true;
    }
    return false;
}

bool IsARealSingularity( double x0, double y0, int chart, int vfindex )
{
    double pcoord[3];
    
    switch( chart )
    {
        case CHART_R2: MATHFUNC(R2_to_sphere)(x0,y0,pcoord); break;
        case CHART_U1: MATHFUNC(U1_to_sphere)(x0,y0,pcoord); break;
        case CHART_V1: MATHFUNC(V1_to_sphere)(x0,y0,pcoord); break;
        case CHART_U2: MATHFUNC(U2_to_sphere)(x0,y0,pcoord); break;
        case CHART_V2: MATHFUNC(V2_to_sphere)(x0,y0,pcoord); break;
        default:
            return false;
    }
    
    return IsARealSingularity(pcoord,vfindex);
}

// ---------------------------------------------------------------------
//                      SPHEREDISTANCE
// ---------------------------------------------------------------------
//
// Calculates somehow the distance between two points on the sphere

static double PSphereDistance( double * p, double * q )
{
    double theta1, theta2;
    double rp[2], rq[2];
    double dist;

    if( p[2] == 0 )
    {
        if( q[2] != 0 )
            return 1.0;         // one is finite, the other is infinite
            
        theta1 = atan2( p[1], p[0] );
        theta2 = atan2( q[1], q[0] );
        
        dist = fabs(theta1 - theta2);
        if( fabs(theta1 - theta2 - TWOPI) < dist )
            dist = fabs(theta1 - theta2 - TWOPI);
        if( fabs(theta1 - theta2 + TWOPI) < dist )
            dist = fabs(theta1 - theta2 + TWOPI);

        return dist;
    }
    
    if( q[2] == 0 )
        return 1.0;             // one is finite, the other is infinite
    
    // both are finite:
    
    psphere_to_R2( p[0], p[1], p[2], rp );
    psphere_to_R2( q[0], q[1], q[2], rq );
    
    dist = (rp[0]-rq[0])*(rp[0]-rq[0]) + (rp[1]-rq[1])*(rp[1]-rq[1]);
    return sqrt(dist);
}

static double PLSphereDistance( double * p, double * q )
{
    double theta1, theta2;
    double rp[2], rq[2];
    double dist;

    if( p[0] == 1 && p[1] == 0 )
    {
        if( !(q[0] == 1 && q[1] == 0) )
            return 1.0;         // one is finite, the other is infinite
            
        theta1 = p[2];
        theta2 = q[2];
        
        dist = fabs(theta1 - theta2);
        if( fabs(theta1 - theta2 - TWOPI) < dist )
            dist = fabs(theta1 - theta2 - TWOPI);
        if( fabs(theta1 - theta2 + TWOPI) < dist )
            dist = fabs(theta1 - theta2 + TWOPI);

        return dist;
    }
    
    if( q[0] == 1 && q[1] == 0 )
        return 1.0;             // one is finite, the other is infinite
    
    // both are finite:
    
    plsphere_to_R2( p[0], p[1], p[2], rp );
    plsphere_to_R2( q[0], q[1], q[2], rq );
    
    dist = (rp[0]-rq[0])*(rp[0]-rq[0]) + (rp[1]-rq[1])*(rp[1]-rq[1]);
    return sqrt(dist);    
}

// ---------------------------------------------------------------------
//                      MARKSINGULARITY
// ---------------------------------------------------------------------

struct positionitem
{
    double pcoord[3];
    struct genericsingularity * s;
};

static void MarkSingularity( struct genericsingularity * s, struct positionitem * * plist, int * numpos, int vfindex, bool plweights )
{
    struct positionitem * positions;
    double pcoord[3];
    int N,k;
    
    switch( s->chart )
    {
        case CHART_R2: MATHFUNC(R2_to_sphere)(s->x0,s->y0,pcoord); break;
        case CHART_U1: MATHFUNC(U1_to_sphere)(s->x0,s->y0,pcoord); break;
        case CHART_V1: MATHFUNC(V1_to_sphere)(s->x0,s->y0,pcoord); break;
        case CHART_U2: MATHFUNC(U2_to_sphere)(s->x0,s->y0,pcoord); break;
        case CHART_V2: MATHFUNC(V2_to_sphere)(s->x0,s->y0,pcoord); break;
        default:
            return;
    }

//    Debug( "Sing (x0,y0)=(%f,%f) Chart=%d P=(%f,%f,%f)", s->x0,s->y0,s->chart,pcoord[0],pcoord[1],pcoord[2]);

    if( !IsARealSingularity( pcoord, vfindex ) )
    {
        s->position = POSITION_VIRTUAL;
        return;
    }

    N = *numpos;
    positions = *plist;
    
    for( k = 0; k < N; k++ )
    {
        if( plweights )
        {
            if( PLSphereDistance( pcoord, positions[k].pcoord ) < 1e-8 )
            {
//                Debug( "coinciding" );
                s->position = POSITION_COINCIDING;
                positions[k].s->position = POSITION_COINCIDING_MAIN;
                return;
            }
        }
        else
        {
            if( PSphereDistance( pcoord, positions[k].pcoord ) < 1e-8 )
            {
//                Debug( "coinciding" );
                s->position = POSITION_COINCIDING;
                positions[k].s->position = POSITION_COINCIDING_MAIN;
                return;
            }
        }
    }
//    Debug( "standalone" );
    s->position = POSITION_STANDALONE;
    
    positions = (struct positionitem *)realloc( positions, sizeof(struct positionitem) * (N+1) );
    positions[N].pcoord[0] = pcoord[0];
    positions[N].pcoord[1] = pcoord[1];
    positions[N].pcoord[2] = pcoord[2];
    positions[N].s = s;
    
    *plist = positions;
    *numpos = N+1;
}

// ---------------------------------------------------------------------
//                      EXAMINEPOSITIONSOFSINGULARITIES
// ---------------------------------------------------------------------

void QPVFStudy::ExaminePositionsOfSingularities( void )
{
    int i;
    struct genericsingularity * s;
    struct positionitem * positions;
    int numpositions;
    
    positions = NULL;
    numpositions = 0;

    for( i = 0; i < ThisVF->numvf; i++ )
    {
        s = (struct genericsingularity *)(vf[i]->first_saddle_point);
        while( s != NULL )
        {
            MarkSingularity( s, &positions, &numpositions, i, plweights );
            s = s->next;
        }
        s = (struct genericsingularity *)(vf[i]->first_se_point);
        while( s != NULL )
        {
            MarkSingularity( s, &positions, &numpositions, i, plweights );
            s = s->next;
        }
        s = (struct genericsingularity *)(vf[i]->first_node_point);
        while( s != NULL )
        {
            MarkSingularity( s, &positions, &numpositions, i, plweights );
            s = s->next;
        }
        s = (struct genericsingularity *)(vf[i]->first_sf_point);
        while( s != NULL )
        {
            MarkSingularity( s, &positions, &numpositions, i, plweights );
            s = s->next;
        }
        s = (struct genericsingularity *)(vf[i]->first_wf_point);
        while( s != NULL )
        {
            MarkSingularity( s, &positions, &numpositions, i, plweights );
            s = s->next;
        }
        s = (struct genericsingularity *)(vf[i]->first_de_point);
        while( s != NULL )
        {
            MarkSingularity( s, &positions, &numpositions, i, plweights );
            s = s->next;
        }
    }
    if( positions != NULL )
        free( positions );
}
