#define TABLE_H

class QTextEdit;

// -----------------------------------------------------------------------
//                      General polynomial expressions
// -----------------------------------------------------------------------

// Linked list of univariate terms a*x^i

struct term1
{
    int exp;
    double coeff;
    struct term1 * next_term1;
};

typedef struct term1 * P4POLYNOM1;

// Linked list of terms a*x^i*y^j

struct term2
{
    int exp_x;
    int exp_y;
    double coeff;
    struct term2 * next_term2;
};

// Linked list of terms a*r^i*cos(theta)^j*sin(theta)^k

typedef struct term2 * P4POLYNOM2;

struct term3
{
    int exp_r;
    int exp_Co;
    int exp_Si;
    double coeff;
    struct term3 * next_term3;
};

typedef struct term3 * P4POLYNOM3;

// -----------------------------------------------------------------------
//                              Orbits
// -----------------------------------------------------------------------

struct orbits_points
{
    int color;          // color of seperatrice

    double pcoord[3];   // point on the poincare sphere -> p=(X,Y,Z)
                        // or 
                        // on the poincare-lyapunov sphere 
                        // -> p=(0,x,y) or p=(1,r,theta)
    int dashes;
    int dir;            // if we have a line of sing at infinity and have to change
    int type;           // the direction if we integrate the orbit of separatrice 
                        // and sometimes the type

    struct orbits_points * nextpt;
};

typedef struct orbits_points * P4ORBIT;

struct orbits
{
    double pcoord[3];               // startpoint
    int color;
    P4ORBIT firstpt;                // orbit
    P4ORBIT currentpt;              // orbit
    struct orbits * next;           // linked list to new orbit
};

// -----------------------------------------------------------------------
//                      Blow up structure
// -----------------------------------------------------------------------

struct transformations
{
    double x0, y0;                      // translation point
    int c1,c2, d1,d2,d3,d4;             // F(x,y)=(c1*x^d1*y^d2,c2*x^d3*y^d4)
    int d;                              // X/x^d
    struct transformations * next_trans;
};

struct blow_up_points
{
    int n;                              // number of transformations
    struct transformations * trans;
    double x0,y0;                       // last point that is not degenerate
    double a11,a12,a21,a22;             // transformation matrix
    struct term2 * vector_field[2];     // vector field
    struct term1 * sep;                 // sep (t,g(t))
    int type;                           // type of seperatrice (STYPE_STABLE, ...)
    bool integrating_in_local_chart;    // if true then use the blow up vector field if 
                                        // the modulus of the last point of the separatrices is less
                                        // than 1
    double point[2];                    // end point sep in blow up chart

    struct orbits_points * first_sep_point;
    struct orbits_points * last_sep_point;
    struct blow_up_points * next_blow_up_point;
};

struct sep
{
    struct orbits_points * first_sep_point;
    struct orbits_points * last_sep_point;
    int type;                           // STYPE_STABLE, UNSTABLE, CENSTABLE or CENUNSTABLE
    int direction;
    int d;                              // if d=0 then separatrix is of the form (t,f(t)), otherwhise (f(t),t).
    bool notadummy;                     // 0 if separatrice is a copy of a structure (obtained
                                        // through a symmetry)
    struct term1 * separatrice;         // if d=0 -> (t,f(t)), d=1 ->(f(t),t)
    struct sep * next_sep;
};

// -----------------------------------------------------------------------
//                          Singularities
// -----------------------------------------------------------------------

// in a piecewise system, singularities may coincide along the bifurcation
// lines.  In that case, we mark the singularity in a different way.
// Of course, on screen it need only be marked once and not several times.

#define POSITION_VIRTUAL                0       // virtual singularity
#define POSITION_STANDALONE             1       // stand-alone singularity
#define POSITION_COINCIDING_MAIN        2       // singularity coincides and will be drawn
#define POSITION_COINCIDING_VIRTUAL     3       // is a virtual singularity, but coincides with a real one so do not plot at all
#define POSITION_COINCIDING             4       // singularity coincides, but is already drawn 

struct genericsingularity       // part of the structure that is the same for all types
{
    double x0;
    double y0;
    struct genericsingularity * next;
    int chart;
    int position;               // POSITION_ constants
};

struct saddle
{
    double x0;
    double y0;
    struct saddle * next_saddle;
    int chart;
    int position;               // POSITION_ constants

    double epsilon;
    bool notadummy;

    struct sep * separatrices;
    struct term2 * vector_field[2];     // vector field {xdot,ydot}
    double a11,a12,a21,a22;             // transformation matrix
};

struct semi_elementary
{
    double x0;
    double y0;
    struct semi_elementary * next_se;
    int chart;
    int position;               // POSITION_ constants

    double epsilon;
    bool notadummy;
    
    struct sep * separatrices;          // center sep (t,f(t)), sep (g(t),t)
    struct term2 * vector_field[2];     // vector field
    double a11,a12,a21,a22;             // transformation matrix

    int type;                           // type of semi-elementary point
};

struct degenerate
{
    double x0;
    double y0;
    struct degenerate * next_de;
    int chart;
    int position;               // POSITION_ constants
    
    double epsilon;
    int notadummy;

    struct blow_up_points * blow_up;
};

struct node
{
    double x0;
    double y0;
    struct node * next_node;
    int chart;
    int position;               // POSITION_ constants

    int stable;
};

struct strong_focus
{
    double x0;
    double y0;
    struct strong_focus * next_sf;
    int chart;
    int position;               // POSITION_ constants

    int stable;
};

struct weak_focus
{
    double x0;
    double y0;
    struct weak_focus * next_wf;
    int chart;
    int position;               // POSITION_ constants

    int type;
};


// -----------------------------------------------------------------------
//                          Some definitions
// -----------------------------------------------------------------------

#define CHART_R2            0
#define CHART_U1            1
#define CHART_U2            2
#define CHART_V1            3
#define CHART_V2            4

#define SADDLE              1
#define NODE                2
#define WEAK_FOCUS          3
#define STRONG_FOCUS        4
#define SEMI_HYPERBOLIC     5  
#define NON_ELEMENTARY      6 

#define STYPE_UNSTABLE      (1)     // separatrice type
#define STYPE_STABLE        (-1)
#define STYPE_CENUNSTABLE   2
#define STYPE_CENSTABLE     (-2)
#define STYPE_ORBIT         3

#define SINGTYPE_UNSTABLE   STYPE_UNSTABLE
#define SINGTYPE_STABLE     STYPE_STABLE
#define SINGTYPE_CENTER     4

#define OT_STABLE           STYPE_STABLE
#define OT_UNSTABLE         STYPE_UNSTABLE
#define OT_CENT_STABLE      STYPE_CENSTABLE
#define OT_CENT_UNSTABLE    STYPE_CENUNSTABLE
#define OT_ORBIT            STYPE_ORBIT

#define TYPEOFVIEW_PLANE    0
#define TYPEOFVIEW_SPHERE   1
#define TYPEOFVIEW_U1       2
#define TYPEOFVIEW_U2       3
#define TYPEOFVIEW_V1       4
#define TYPEOFVIEW_V2       5
#define TYPEOFVIEW_U1U2     6
#define TYPEOFVIEW_V1V2     7

class QVFStudy;

typedef struct _VFRegionResult
{
    int vfindex;        // which vector field
    int * signs;
} VFREGIONRESULT;

typedef struct _CurveRegionResult
{
    int  curveindex;        // which curve are we talking about
    int * signs;
} CURVEREGIONRESULT;

typedef struct _CurveResult
{
    P4POLYNOM2                  sep;
    P4POLYNOM2                  sep_U1;
    P4POLYNOM2                  sep_U2;
    P4POLYNOM2                  sep_V1;
    P4POLYNOM2                  sep_V2;
    P4POLYNOM3                  sep_C;
    struct orbits_points *      sep_points;
} CURVERESULT;

class QPVFStudy : public QObject
{
public:
    QVFStudy * * vf;
    QVFStudy * vfK;                     // shortcut for vf[K]. Must be updated whenever K changes
    int K;                              // K will be throughout the current vector field selected
    
//    int numvf;                        // LOADED FROM THISVF
//    int numvfregions;
//    int numcurveregions;
    CURVERESULT * curves;    
//    VFREGIONRESULT * vfregions;
//    CURVEREGIONRESULT * curveregions;

    int typeofstudy;
    int typeofview;                     // TYPEOFVIEW_PLANE or TYPEOFVIEW_SPHERE
    bool plotvirtualsingularities;      // true or false
    int p;
    int q;
    bool plweights;                     // true if p<>1 or q<>1; false if p=q=1
    double config_projection;
  
    double double_p;                    // shortcuts: = (double)p
    double double_q;                    // = (double)q
    double double_p_plus_q;             // = (double)(p+q)
    double double_p_minus_1;            // = (double)(p-1)
    double double_q_minus_1;            // = (double)(q-1)
    double double_q_minus_p;            // = (double)(q-p)
    double xmin, xmax, ymin, ymax;      // in case of local study

    // limit cycles and orbits

    struct orbits *             first_lim_cycle;
    struct orbits *             first_orbit;

    // run-time when plotting

    struct orbits *             current_orbit;
    struct orbits *             current_lim_cycle;

    double                      selected_ucoord[2];
    struct saddle *             selected_saddle_point;
    struct semi_elementary *    selected_se_point;
    struct degenerate *         selected_de_point;
    struct sep *                selected_sep;
    struct blow_up_points *     selected_de_sep;
    int                         selected_sep_vfindex;

    double                      config_currentstep;
    bool                        config_dashes;
    bool                        config_kindvf;              // true for original VF, false for reduced
    int                         config_lc_value;
    int                         config_lc_numpoints;
    double                      config_hma;
    double                      config_hmi;
    double                      config_branchhmi;
    double                      config_step;
    double                      config_tolerance;
    int                         config_intpoints;

public:
    QPVFStudy();                 // constructor
    ~QPVFStudy();                // destructor

    bool ReadTables( QString, bool, bool );
    void ExaminePositionsOfSingularities( void );
    bool ReadPiecewiseData( FILE * );
    void Dump( QString basename );
    void Reset( void );
    void SetupCoordinateTransformations( void );    // see math_p4.cpp
    bool ReadSeparatingCurve( FILE *, CURVERESULT * );

    void ResetCurveInfo( int );

    // coordinate transformation routines, set up when starting the plot

    void (*viewcoord_to_sphere)( double, double, double * );
    bool (*sphere_to_viewcoordpair)( double *, double *, double *, double *, double *, double * );

    void (*finite_to_viewcoord)( double, double, double * );
    void (*sphere_to_viewcoord)( double, double, double, double * );
    bool (*is_valid_viewcoord)( double, double, double * );
    void (*integrate_sphere_sep)( double, double, double, double *,
                                double *, int *, int *, int *, int *, double, double );
    void (*U1_to_sphere)( double, double, double * );
    void (*U2_to_sphere)( double, double, double * );
    void (*V1_to_sphere)( double, double, double * );
    void (*V2_to_sphere)( double, double, double * );

    void (*sphere_to_U1)( double, double, double, double * );
    void (*sphere_to_U2)( double, double, double, double * );
    void (*sphere_to_V1)( double, double, double, double * );
    void (*sphere_to_V2)( double, double, double, double * );

    void (*sphere_to_R2)( double, double, double, double * );
    void (*R2_to_sphere)( double, double, double * );
    void (*integrate_sphere_orbit)( double, double, double, double *,
                                double *, int *, int *, double, double );
    double (*eval_lc)( double *, double, double, double );
    bool (*less2)( double *, double * );
    int (*change_dir)( double * );
};

// -----------------------------------------------------------------------
//                          Results class
// -----------------------------------------------------------------------

class QVFStudy : public QObject
{
public:
    // general information

    bool singinf;
    int dir_vec_field;
    QPVFStudy * parent;

    // vector field in various charts

    P4POLYNOM2                  f_vec_field[2];
    P4POLYNOM2                  vec_field_U1[2];
    P4POLYNOM2                  vec_field_U2[2];
    P4POLYNOM2                  vec_field_V1[2];
    P4POLYNOM2                  vec_field_V2[2];
    P4POLYNOM3                  vec_field_C[2];

    // singular points and their properties:

    struct saddle *             first_saddle_point;
    struct semi_elementary *    first_se_point;
    struct node *               first_node_point;
    struct strong_focus *       first_sf_point;
    struct weak_focus *         first_wf_point;
    struct degenerate *         first_de_point;

    // Greatest common factor if present:

    P4POLYNOM2                  gcf;
    P4POLYNOM2                  gcf_U1;
    P4POLYNOM2                  gcf_U2;
    P4POLYNOM2                  gcf_V1;
    P4POLYNOM2                  gcf_V2;
    P4POLYNOM3                  gcf_C;
    struct orbits_points *      gcf_points;

    // ------ Configuration

public:
    QVFStudy();                 // constructor
    ~QVFStudy();                // destructor

public:

    // initialization and destruction of structures

    void Reset( void );

    void DeleteSaddle( struct saddle * );
    void DeleteSemiElementary( struct semi_elementary * );
    void DeleteNode( struct node * );
    void DeleteStrongFocus( struct strong_focus * );
    void DeleteWeakFocus( struct weak_focus * );
    void DeleteDegenerate( struct degenerate * );
    void DeleteSeparatrices( struct sep * );
    void DeleteTransformations( struct transformations * );
    void DeleteBlowup( struct blow_up_points * b );

    // reading of the Maple/Reduce results

    bool ReadTables( FILE *, FILE *, FILE * );
    bool ReadGCF( FILE * fp );
    bool ReadVectorField( FILE * fp, P4POLYNOM2 * vf );
    bool ReadVectorFieldCylinder( FILE * fp, P4POLYNOM3 * vf );
    bool ReadPoints( FILE * fp );
    bool ReadSaddlePoint( FILE * fp );
    bool ReadSemiElementaryPoint( FILE * fp );
    bool ReadStrongFocusPoint( FILE * fp );
    bool ReadWeakFocusPoint( FILE * fp );
    bool ReadDegeneratePoint( FILE * fp );
    bool ReadNodePoint( FILE * fp );
    bool ReadBlowupPoints( FILE * fp, struct blow_up_points * b, int n );
    bool ReadTransformations( FILE * fp, struct transformations * trans, int n );

    void Dump( QTextEdit * );
};

#define LINESTYLE_DASHES    1
#define LINESTYLE_POINTS    0

void DeleteLimitCycle( struct orbits * );
void DeleteOrbitPoint( P4ORBIT p );
void DeleteOrbit( struct orbits * );
bool ReadTerm1( FILE * fp, struct term1 * p, int N );
bool ReadTerm2( FILE * fp, struct term2 * p, int N );
bool ReadTerm3( FILE * fp, struct term3 * p, int N );
extern double eval_curve( const CURVERESULT *, const double * );
